# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Benoit Pauwels - Stacked data management
#               (e.g. iteration index)

"""
A database of function calls and design variables
*************************************************
"""

from __future__ import division, unicode_literals

import logging
from ast import literal_eval
from hashlib import sha1
from itertools import chain, islice
from typing import (
    Any,
    Callable,
    ItemsView,
    Iterable,
    KeysView,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    ValuesView,
)
from xml.etree.ElementTree import parse as parse_element

import h5py
from numpy import (
    array,
    atleast_2d,
    concatenate,
    float64,
    isclose,
    ndarray,
    string_,
    uint8,
)
from numpy.linalg import norm
from six import string_types

from gemseo.utils.ggobi_export import save_data_arrays_to_xml
from gemseo.utils.hdf5 import get_hdf5_group
from gemseo.utils.py23_compat import OrderedDict

LOGGER = logging.getLogger(__name__)

# Type of the values associated to the keys (values of input variables) in the database
DatabaseValueType = Mapping[str, Union[float, ndarray, List[int]]]
ReturnedHdfMissingOutputType = Tuple[
    Mapping[str, Union[float, ndarray, List[int]]], Union[None, Mapping[str, int]]
]


class Database(object):
    """Class to store evaluations of functions, such as DOE or optimization histories.

    Avoid multiple calls of the same functions,
    useful when simulations are costly.

    It is also used to store inputs and retrieve them
    for optimization graphical post processing and plots
    generation.

    Can be serialized to HDF5 for portability and cold post processing.

    The database is based on a two-levels dictionary-like mapping such as
    ``{key_level_1: {key_level_2: value_level_2} }`` with:
        * ``key_level_1``: the values of the input design variables that have been used
          during the evaluations;
        * ``key_level_2``: the name of the output functions that have been returned,
          the name of the gradient
          (the gradient of a function called ``func`` is typically denoted as ``@func``),
          the iteration numbers which correspond to the evaluation
          (this key is typically denoted ``Iter``)
          and any additional information related to the methods which use the database
          (penalization parameter, trust region radii...);
        * ``value_level_2``: depending on the ``key_level_2``, but is typically a float
          for an output function, an array for a gradient or a list for
          the iterations (when several iterations are stored for a same point,
          this means that the point has been replicated during the process at
          these iterations).

    Attributes:
        name (str): The name of the database.
    """

    missing_value_tag = "NA"
    KEYSSEPARATOR = "__KEYSSEPARATOR__"
    GRAD_TAG = "@"
    ITER_TAG = "Iter"

    def __init__(
        self,
        input_hdf_file=None,  # type: Optional[str]
        name=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """
        Args:
            input_hdf_file: The path to a HDF5 file from which the database is created.
                If None, do not import a database.
            name: The name to be given to the database.
                If None, use the class name.
        """
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

        self.__dict = OrderedDict()
        self.__max_iteration = 0
        self.__store_listeners = []
        self.__newiter_listeners = []

        # This list enables to temporary save the last inputs that have been
        # stored before any export_hdf is called.
        # It is used to append the exported hdf file.
        self.__hdf_export_buffer = []

        if input_hdf_file is not None:
            self.import_hdf(input_hdf_file)

    def __setitem__(
        self,
        key,  # type: Union[ndarray, HashableNdarray]
        value,  # type: DatabaseValueType
    ):  # type: (...) -> None
        """Set an item to the database.

        Args:
            key: The key of the item (values of input variables).
            value: The value of the item (output functions).

        Raises:
            TypeError:
                * If the key is neither an array, nor a hashable array.
                * If the value is not a dictionary.
        """
        if not isinstance(key, (ndarray, HashableNdarray)):
            raise TypeError(
                "Optimization history keys must be design variables numpy arrays."
            )
        if not isinstance(value, dict):
            raise TypeError("Optimization history values must be data dictionary.")
        if isinstance(key, HashableNdarray):
            self.__dict[key] = OrderedDict(value)
        else:
            self.__dict[HashableNdarray(key, True)] = OrderedDict(value)

    def is_new_eval(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> bool
        """Whether storing the given values would generate a new iteration.

        Args:
            x_vect: The design variables values.

        Returns:
            Whether a new iteration would occur after storing the values.
        """
        curr_val = self.get(x_vect)

        if curr_val is None:
            return True

        n_cval = len(curr_val)
        return (n_cval == 1 and self.ITER_TAG in curr_val) or n_cval == 0

    @staticmethod
    def __get_hashed_key(
        x_vect,  # type: Union[ndarray, HashableNdarray]
    ):  # type: (...) -> HashableNdarray
        """Convert an array to a hashable array.

        This array basically represent a key of the first level of the database.

        Args:
            x_vect: An array.

        Returns:
            The input array converted to a hashable array.

        Raises:
            TypeError: If the input is not an array.
        """
        if not isinstance(x_vect, (ndarray, HashableNdarray)):
            raise TypeError("Database keys must have ndarray type.")
        if isinstance(x_vect, ndarray):
            return HashableNdarray(x_vect)
        return x_vect

    def __getitem__(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> DatabaseValueType
        """Get an item value from the database.

        Args:
            x_vect: The key of the item.

        Returns:
            The value of the item.
        """
        hashed = self.__get_hashed_key(x_vect)
        return self.__dict[hashed]

    def __delitem__(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> None
        """Delete an item from the database.

        Args:
            x_vect: The key of the item that must be deleted.
        """
        hashed = self.__get_hashed_key(x_vect)
        del self.__dict[hashed]

    def setdefault(
        self,
        key,  # type: ndarray
        default,  # type: DatabaseValueType
    ):  # type: (...) -> DatabaseValueType
        """Set a default database entry.

        Args:
            key: The key of the default item.
            default: The value of the item.

        Returns:
            The value which is set.

        Raises:
            TypeError:
                * If the key is not ndarray type.
                * If the value is not dictionary type.
        """
        if not isinstance(key, (ndarray, HashableNdarray)):
            raise TypeError("Database keys must have ndarray type.")
        if not isinstance(default, dict):
            raise TypeError("Database values must have dictionary type")

        if isinstance(key, ndarray):
            return self.__dict.setdefault(HashableNdarray(key), default)

        return self.__dict.setdefault(key, default)

    def __len__(self):  # type: (...) -> int
        """Get the length of the database.

        Returns:
            The length.
        """
        return len(self.__dict)

    def keys(self):  # type: (...) -> KeysView[ndarray]
        """Database keys generator.

        Yields:
            The next key in the database.
        """
        return self.__dict.keys()

    def values(self):  # type: (...) -> ValuesView[DatabaseValueType]
        """Database values generator.

        Yields:
            The next value in the database.
        """
        return self.__dict.values()

    def items(self):  # type: (...) -> ItemsView[ndarray, DatabaseValueType]
        """Database items generator.

        Yields:
            The next key and value in the database.
        """
        return self.__dict.items()

    def get_value(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> DatabaseValueType
        """Return a value in the database.

        Args:
            x_vect: The key associated to the value.

        Returns:
            The values that correspond to the required key.

        Raises:
            KeyError: If the key does not exist.
        """
        return self[x_vect]

    def get_max_iteration(self):  # type: (...) -> int
        """Return the maximum number of iterations.

        Returns:
            The max number of iterations.
        """
        return self.__max_iteration

    def get_x_history(self):  # type: (...) -> List[ndarray]
        """Return the history of the input design variables ordered by calls.

        Returns:
            This values of input design variables.
        """
        return [x_vect.unwrap() for x_vect in self.__dict.keys()]

    def get_last_n_x(
        self, n  # type: int
    ):  # type: (...) -> List[ndarray]
        """Return the last n ordered calls of the input design variables.

        Args:
            n: The number of last returned calls.

        Returns:
            The values of the input design variables for the last n calls.

        Raises:
            ValueError: If the number n is higher than the size of the database.
        """
        n_max = len(self)
        if n > n_max:
            raise ValueError(
                "The number n = {} must be lower than "
                "the database size = {}".format(n, n_max)
            )
        return [
            x_vect.unwrap() for x_vect in islice(self.__dict.keys(), n_max - n, n_max)
        ]

    def get_index_of(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> int
        """Return the index of an input values in the database.

        Args:
            x_vect: The input values.

        Returns:
            The index of the input values in the database.

        Raises:
            KeyError: If the required key is not found.
        """
        hashed = HashableNdarray(x_vect)

        for i, key in enumerate(self.__dict.keys()):
            if key == hashed:
                return i
        raise KeyError(x_vect)

    def get_x_by_iter(
        self, iteration  # type: int
    ):  # type: (...) -> ndarray
        """Return the values of the input design variables at a specified iteration.

        Args:
            iteration: The required iteration.

        Returns:
            The values of the input design variables.

        Raises:
            ValueError:
                * If the database is empty.
                * If the required iteration is higher than the maximum
                number of iterations in the database.
        """
        nkeys = len(self.__dict)
        if nkeys == 0:
            raise ValueError("The database is empty.")
        if iteration < 0:
            iteration = nkeys + iteration
        if iteration >= nkeys or (iteration < 0 and -iteration > nkeys):
            raise ValueError(
                "Iteration required should be lower "
                "than the maximum number of iterations = {} "
                "got instead = {}".format(len(self) - 1, iteration)
            )

        # The database dictionary uses the input design variables as keys for the
        # function values. Here we convert it to an iterator that returns the
        # key located at the required iteration using the islice method from
        # itertools.
        key = next(islice(iter(self.__dict), iteration, iteration + 1))
        return key.unwrap()

    def clear(self):  # type: (...) -> None
        """Clear the database."""
        self.__dict.clear()

    def clean_from_iterate(
        self, iterate  # type: int
    ):  # type: (...) -> None
        """Delete the iterates after a given iteration number.

        Args:
            iterate: The iteration number.
        """

        def gen_todel():
            for iterate_number, x_vect in enumerate(self.__dict.keys()):
                # remove iterations beyond limit iterate number
                if iterate < iterate_number:
                    yield x_vect

        # Copies only the keys after iterate
        to_del = list(gen_todel())
        for key in to_del:
            del self.__dict[key]
        self.__max_iteration = len(self)

    def remove_empty_entries(self):  # type: (...) -> None
        """Remove items when the key is associated to an empty value."""
        empt = [
            k
            for k, v in self.items()
            if len(v) == 0 or (len(v) == 1 and list(v.keys())[0] == self.ITER_TAG)
        ]
        for k in empt:
            del self[k]

    def filter(
        self, data_list_to_keep  # type: Iterable[str]
    ):  # type: (...) -> None
        """Filter the database so that only the required output functions are kept.

        Args:
            data_list_to_keep: The name of output functions that must be kept.
        """
        data_list_to_keep = set(data_list_to_keep)
        for val in self.values():
            keys_to_del = set(val.keys()) - data_list_to_keep
            for key in keys_to_del:
                del val[key]

    def get_func_history(
        self,
        funcname,  # type: str
        x_hist=False,  # type: bool
    ):  # type: (...) -> Union[ndarray, Tuple[ndarray, List[ndarray]]]
        """Return the history of the output function values.

        This function can also return the history of the input values.

        Args:
            funcname: The name of the function.
            x_hist: Whether the input values history is also returned.

        Returns:
            The function values history.
            The input values history if required.
        """
        outf_l = []
        x_history = []
        for x_vect, out_val in self.items():
            val = out_val.get(funcname)
            if val is not None:
                if isinstance(val, ndarray) and val.size == 1:
                    val = val[0]
                outf_l.append(val)
                if x_hist:
                    x_history.append(x_vect.unwrap())
        outf = array(outf_l)
        if x_hist:
            return outf, x_history

        return outf

    def get_func_grad_history(
        self,
        funcname,  # type: str
        x_hist=False,  # type: bool
    ):  # type: (...) -> Union[ndarray, Tuple[ndarray, List[ndarray]]]
        """Return the history of the gradient values of any function.

        The function can also return the history of the input values.

        Args:
            funcname: The name of the function.
            x_hist: Whether the input values history is also returned.

        Returns:
            The gradient values history of the function.
            The input values history if required.
        """
        gradient_name = self.get_gradient_name(funcname)
        return self.get_func_history(funcname=gradient_name, x_hist=x_hist)

    def is_func_grad_history_empty(
        self, funcname  # type: str
    ):  # type: (...) -> bool
        """Check if the history of the gradient of any function is empty.

        Args:
            funcname: The name of the function.

        Returns:
            True if the history of the gradient is empty, False otherwise.
        """
        return len(self.get_func_grad_history(funcname, x_hist=False)) == 0

    def contains_x(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> bool
        """Check if the history contains a specific value of input design variables.

        Args:
            x_vect: The input values that is checked.

        Returns:
            True is the input values are in the history, False otherwise.
        """
        return HashableNdarray(x_vect) in self.__dict

    def get_f_of_x(
        self,
        fname,  # type: str
        x_vect,  # type: ndarray
        dist_tol=0.0,  # type: float
    ):  # type: (...) -> Optional[float, ndarray, List[int]]
        """Return the output function values of any input values in the database.

        Args:
            fname: The name of the required output function.
            x_vect: The required input values.
            dist_tol: In the N-dimensional space of input design variables,
                this value is the threshold of the normalized distance
                with respect to the required point (prescribed input values)
                below which any point in the database can be selected.

        Returns:
            The values of the output function corresponding to
            the required input values.
            ``None`` if no point matches.
        """
        if isclose(dist_tol, 0.0, rtol=1e-16):
            vals = self.get(x_vect)
            if vals is not None:
                return vals.get(fname)  # Returns None if not in self
        else:
            for x_key, vals in self.items():
                x_v = x_key.unwrap()
                if norm(x_v - x_vect) <= dist_tol * norm(x_v):
                    return vals.get(fname)
        return None

    def get(
        self,
        x_vect,  # type: ndarray
        default=None,  # type: Any
    ):  # type: (...) -> Any
        """Return the value of the required key if the key is in the dictionary.

        Args:
            x_vect: The required key.
            default: The values that is returned if the key is not found.

        Returns:
            The value corresponding to the key or the prescribed default value
            if the key is not found.

        Raises:
            TypeError: If the type of the required key is neither an array
                nor a hashable array.
        """
        if not isinstance(x_vect, (HashableNdarray, ndarray)):
            raise TypeError(
                "The key must be an ndarray with the values "
                "of the input design variables."
            )
        if isinstance(x_vect, ndarray):
            x_vect = HashableNdarray(x_vect)
        return self.__dict.get(x_vect, default)

    def pop(
        self, key  # type: ndarray
    ):  # type: (...) -> DatabaseValueType
        """Remove the required key from the database and return the corresponding value.

        Args:
            key: The required key.

        Returns:
            The values corresponding to the key.

        Raises:
            KeyError: If the key is not found.
        """
        return self.__dict.pop(key)

    def contains_dataname(
        self,
        data_name,  # type: str
        skip_grad=False,  # type: bool
    ):  # type: (...) -> bool
        """Check if the database has an output function with the required name.

        Args:
            data_name: The required name of the output function.
            skip_grad: True if the name of the gradients are skipped during the search,
                False otherwise.

        Returns:
            Whether the required output function name is in the database.
        """
        return data_name in self.get_all_data_names(skip_grad=skip_grad)

    def store(
        self,
        x_vect,  # type: ndarray
        values_dict,  # type: DatabaseValueType
        add_iter=True,  # type: bool
    ):  # type: (...) -> None
        """Store the output values associated to the input values.

        Args:
            x_vect: The input values.
            values_dict: The output values corresponding to the input values.
            add_iter: True if iterations are added to the output values, False otherwise.
        """
        self.__hdf_export_buffer.append(self.__get_hashed_key(x_vect))

        n_values = len(values_dict)
        values_not_empty = n_values > 1 or (
            n_values == 1 and self.ITER_TAG not in values_dict
        )
        if self.contains_x(x_vect):
            curr_val = self.get_value(x_vect)
            # No new keys = already computed = new iteration
            # otherwise just calls to other functions
            cval_ok = (len(curr_val) == 1 and self.ITER_TAG in curr_val) or not curr_val
            curr_val.update(OrderedDict(values_dict))
            self[x_vect] = curr_val
            self.notify_store_listeners()
            # Notify the new iteration after storing x_vect
            # because callbacks may need an updated x_vect
            if cval_ok and values_not_empty:
                self.notify_newiter_listeners(x_vect)
        else:
            self.__max_iteration += 1
            if add_iter:
                values_dict = dict(
                    values_dict, **{self.ITER_TAG: [self.__max_iteration]}
                )
            self[x_vect] = values_dict
            self.notify_store_listeners()
            if values_not_empty:
                self.notify_newiter_listeners(x_vect)

    def add_store_listener(
        self,
        listener_func,  # type: Callable
    ):  # type: (...) -> None
        """Add a listener to be called when an item is stored to the database.

        Args:
            listener_func: The function to be called.

        Raises:
            TypeError: If the argument is not a callable
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable")
        self.__store_listeners.append(listener_func)

    def add_new_iter_listener(
        self,
        listener_func,  # type: Callable
    ):  # type: (...) -> None
        """Add a listener to be called when a new iteration is stored to the database.

        Args:
            listener_func: The function to be called.

        Raises:
            TypeError: If the argument is not a callable.
        """
        if not callable(listener_func):
            raise TypeError("Listener function is not callable.")
        self.__newiter_listeners.append(listener_func)

    def clear_listeners(self):  # type: (...) -> None
        """Clear all the listeners."""
        self.__store_listeners = []
        self.__newiter_listeners = []

    def notify_store_listeners(self):  # type: (...) -> None
        """Notify the listeners that a new entry was stored in the database."""
        for func in self.__store_listeners:
            func()

    def notify_newiter_listeners(
        self,
        x_vect=None,  # type: Optional[ndarray]
    ):  # type: (...) -> None
        """Notify the listeners that a new iteration is ongoing.

        Args:
            x_vect: The values of the design variables. If None, use
                the values of the last iteration.
        """
        if x_vect is None:
            x_vect = self.get_x_by_iter(-1)

        for func in self.__newiter_listeners:
            try:
                func(x_vect)
            except TypeError:
                func()

    def get_all_data_names(
        self,
        skip_grad=True,  # type: bool
        skip_iter=False,  # type: bool
    ):  # type: (...) -> Set[str]
        """Return all the names of the output functions contained in the database.

        Args:
            skip_grad: True if the names of the keys corresponding to the gradients
                are not returned, False otherwise.
            skip_iter: True if the names of the keys corresponding to the iteration
                numbers are not returned, False otherwise.

        Returns:
            The names of the output functions.
        """
        names = set()
        for value in self.__dict.values():
            for key in value.keys():
                if skip_grad and key.startswith(self.GRAD_TAG):
                    continue
                names.add(key)
        if skip_iter and self.ITER_TAG in names:
            names.remove(self.ITER_TAG)
        return sorted(names)

    def _format_history_names(
        self,
        functions,  # type: Iterable[str]
        stacked_data,  # type: Iterable[str]
    ):  # type: (...) -> Tuple[Iterable[str], List[str]]
        """Format the names of the output functions to be displayed in the history.

        Args:
            functions: The names of output functions.
            stacked_data: The names of outputs that are stored as Iterable
                in the database (e.g. iterations) and that must be unstacked.

        Returns:
            The names of the functions.
            The names of the stacked outputs.

        Raises:
            ValueError: If the names of stacked data are not a subset of the names
                of the functions.
        """
        if functions is None:
            functions = self.get_all_data_names()
        if stacked_data is None:
            if self.ITER_TAG in functions:
                stacked_data = [self.ITER_TAG]
            else:
                stacked_data = []
        elif not set(stacked_data).issubset(functions):
            raise ValueError(
                "The names of the data to be unstacked ({}) "
                "must be included in the names of the data "
                "to be returned ({}).".format(stacked_data, functions)
            )
        elif self.ITER_TAG in functions and self.ITER_TAG not in stacked_data:
            stacked_data.append(self.ITER_TAG)
        return functions, stacked_data

    def get_complete_history(
        self,
        functions=None,  # type: Optional[Iterable[str]]
        add_missing_tag=False,  # type: bool
        missing_tag="NA",  # type: str
        all_iterations=False,  # type: bool
        stacked_data=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> Tuple[List[List[Union[float, ndarray]]], List[ndarray]]
        """Return the complete history of the optimization: design variables, functions
        and gradients.

        Args:
            functions: The names of output functions.
            add_missing_tag: If True, the tag specified in ``missing_tag`` is added to
                iterations where data are not available.
            missing_tag: The tag that is written when data are missing.
            all_iterations: If True, the points which are called at several
                iterations will be duplicated in the history (each duplicate
                corresponding to a different calling index); otherwise each point
                will appear only once (with the latest calling index)
            stacked_data: The names of outputs corresponding to data stored as
                ``iterable`` (e.g. iterations, penalization parameters or
                trust region radii).

        Returns:
            The history of the output values.
            The history of the input values.
        """
        functions, stacked_data = self._format_history_names(functions, stacked_data)
        f_history = []
        x_history = []
        for x_vect, out_val in self.items():
            # If duplicates are not to be considered, or if no iteration index
            # is specified, then only one entry (the last one) will be written:
            if not all_iterations or self.ITER_TAG not in out_val:
                first_index = -1
                last_index = -1
            # Otherwise all the entries will be written:
            else:
                first_index = 0
                # N.B. if the list of indexes is empty, then no entry will be
                # written.
                last_index = len(out_val[self.ITER_TAG]) - 1
            # Add an element to the history for each duplicate required:
            for duplicate_ind in range(first_index, last_index + 1):
                out_vals = []
                for funcname in functions:
                    if funcname in out_val:
                        if funcname not in stacked_data:
                            out_vals.append(out_val[funcname])
                        # If the data 'funcname' is stacked and there remains
                        # entries to unstack, then unstack the next entry:
                        elif duplicate_ind < len(out_val[funcname]):
                            val = out_val[funcname][duplicate_ind]
                            out_vals.append(val)
                    elif add_missing_tag:
                        out_vals.append(missing_tag)
                if out_vals:
                    f_history.append(out_vals)
                    x_history.append(x_vect.unwrap())
        return f_history, x_history

    @staticmethod
    def __to_real(
        data,  # type: Union[ndarray, Iterable[complex]]
    ):  # type: (...) -> ndarray
        """Convert complex to real numpy array.

        Args:
            data: The input values.

        Returns:
            The real output values.
        """
        return array(array(data, copy=False).real, dtype=float64)

    def _add_hdf_input_dataset(
        self,
        index_dataset,  # type: int
        design_vars_group,  # type: h5py.Group
        design_vars_values,  # type: HashableNdarray
    ):  # type: (...) -> None
        """Add a new input to the hdf group of input values.

        Args:
            index_dataset: The index of the new hdf5 entry.
            design_vars_group: The hdf5 group of the design variable values.
            design_vars_values: The values of the input.

        Raises:
            ValueError: If the dataset name ``index_dataset`` already exists
                in the group of design variables.
        """
        if str(index_dataset) in design_vars_group:
            raise ValueError(
                "Dataset name '{}' already exists in the group "
                "of design variables.".format(index_dataset)
            )

        design_vars_group.create_dataset(
            str(index_dataset), data=design_vars_values.unwrap()
        )

    def _add_hdf_output_dataset(
        self,
        index_dataset,  # type: int
        keys_group,  # type: h5py.Group
        values_group,  # type: h5py.Group
        output_values,  # type: Mapping[str, Union[float, ndarray, List]]
        output_name_to_idx=None,  # type: Optional[Mapping[str, int]]
    ):  # type: (...) -> None
        """Add new outputs to the hdf group of output values.

        Args:
            index_dataset: The index of the new hdf5 entry.
            keys_group: The hdf5 group of the output names.
            values_group: The hdf5 group of the output values.
            output_values: The output values.
            output_name_to_idx: The indices of the output names in ``output_values``.
                If ``None``, these indices are automatically built using the
                order of the names in ``output_values``.
                These indices are used to build the dataset of output vectors.
        """
        self._add_hdf_name_output(index_dataset, keys_group, list(output_values.keys()))

        if not output_name_to_idx:
            output_name_to_idx = dict(zip(output_values, range(len(output_values))))

        # We separate scalar data from vector data in the hdf file.
        # Scalar data are first stored into a list (``values``),
        # then added to the hdf file.
        # Vector data are directly added to the hdf file.
        values = []
        for name, value in output_values.items():
            idx_value = output_name_to_idx[name]
            if isinstance(value, (ndarray, list)):
                self._add_hdf_vector_output(
                    index_dataset, idx_value, values_group, value
                )
            else:
                values.append(value)

        if values:
            self._add_hdf_scalar_output(index_dataset, values_group, values)

    def _get_missing_hdf_output_dataset(
        self,
        index_dataset,  # type: int
        keys_group,  # type: h5py.Group
        output_values,  # type: Mapping[str, Union[float, ndarray, List[int]]]
    ):  # type: (...) -> ReturnedHdfMissingOutputType
        """Return the missing values in the hdf group of the output names.

        Compare the keys of ``output_values`` with the existing names
        in the group of the output names ``keys_group`` in order to know which
        outputs are missing.

        Args:
            index_dataset: The index of the new hdf5 entry.
            keys_group: The hdf5 group of the output names.
            output_values: The output values to be compared with.

        Returns:
            The missing values.
            The indices of the missing outputs.

        Raises:
            ValueError: If the index of the dataset does not correspond to
                an existing dataset.
        """
        name = str(index_dataset)
        if name not in keys_group:
            raise ValueError("The dataset named '{}' does not exist.".format(name))

        existing_output_names = set(out.decode() for out in keys_group[name])
        all_output_names = set(output_values)
        missing_names = all_output_names - existing_output_names

        if not missing_names:
            return {}, None

        missing_names_values = {name: output_values[name] for name in missing_names}
        all_output_idx_mapping = dict(zip(output_values, range(len(output_values))))
        missing_names_idx_mapping = {
            name: all_output_idx_mapping[name] for name in missing_names
        }

        return missing_names_values, missing_names_idx_mapping

    def _add_hdf_name_output(
        self,
        index_dataset,  # type: int
        keys_group,  # type: h5py.Group
        keys,  # type: List[str]
    ):
        """Add new output names to the hdf5 group of output names.

        Create a dataset in the group of output names
        if the dataset index is not found in the group.
        If the dataset already exists, the new names are appended
        to the existing dataset.

        Args:
            index_dataset: The index of the new hdf5 entry.
            keys_group: The hdf5 group of the output names.
            keys: The names that must be added.
        """
        name = str(index_dataset)
        keys = array(keys, dtype=string_)
        if name not in keys_group:
            keys_group.create_dataset(
                name, data=keys, maxshape=(None,), dtype=h5py.string_dtype()
            )
        else:
            offset = len(keys_group[name])
            keys_group[name].resize((offset + len(keys),))
            keys_group[name][offset:] = keys

    def _add_hdf_scalar_output(
        self,
        index_dataset,  # type: int
        values_group,  # type: h5py.Group
        values,  # type: List[float]
    ):  # type: (...) -> None
        """Add new scalar values to the hdf5 group of output values.

        Create a dataset in the group of output values
        if the dataset index is not found in the group.
        If the dataset already exists, the new values are appended to
        the existing dataset.

        Args:
            index_dataset: The index of the new hdf5 entry.
            values_group: The hdf5 group of the output values.
            values: The scalar values that must be added.
        """
        name = str(index_dataset)
        if name not in values_group:
            values_group.create_dataset(
                name, data=self.__to_real(values), maxshape=(None,), dtype=float64
            )
        else:
            offset = len(values_group[name])
            values_group[name].resize((offset + len(values),))
            values_group[name][offset:] = self.__to_real(values)

    def _add_hdf_vector_output(
        self,
        index_dataset,  # type: int
        idx_sub_group,  # type: int
        values_group,  # type: h5py.Group
        value,  # type: Union[ndarray, List[int]]
    ):  # type: (...) -> None
        """Add a new vector of values to the hdf5 group of output values.

        Create a sub-group dedicated to vectors in the group of output
        values.
        Inside this sub-group, a new dataset is created for each vector.
        If the sub-group already exists, it is just appended.
        Otherwise, the sub-group is created.

        Args:
            index_dataset: The index of the hdf5 entry.
            idx_sub_group: The index of the dataset in the sub-group of vectors.
            values_group: The hdf5 group of the output values.
            value: The vector which is added to the group.

        Raises:
            ValueError: If the index of the dataset in the sub-group of vectors
                already exist.
        """
        sub_group_name = "arr_{}".format(index_dataset)

        if sub_group_name not in values_group:
            sub_group = values_group.require_group(sub_group_name)
        else:
            sub_group = values_group[sub_group_name]

        if str(idx_sub_group) in sub_group:
            raise ValueError(
                "Dataset name '{}' already exists in the sub-group of "
                "array output '{}'.".format(idx_sub_group, sub_group_name)
            )

        sub_group.create_dataset(
            str(idx_sub_group), data=self.__to_real(value), dtype=float64
        )

    def _append_hdf_output(
        self,
        index_dataset,  # type: int
        keys_group,  # type: h5py.Group
        values_group,  # type: h5py.Group
        output_values,  # type: Mapping[str, Union[float, ndarray, List[int]]]
    ):  # type: (...) -> None
        """Append the existing hdf5 datasets of the outputs with new values.

        Find the values among ``output_values`` that do not
        exist in the hdf5 datasets and append them to the datasets.

        Args:
            index_dataset: The index of the existing hdf5 entry.
            keys_group: The hdf5 group of the output names.
            values_group: The hdf5 group of the output values.
            output_values: The output values. Only the values which
                do not exist in the dataset will be appended.
        """
        added_values, mapping_to_idx = self._get_missing_hdf_output_dataset(
            index_dataset, keys_group, output_values
        )
        if added_values:
            self._add_hdf_output_dataset(
                index_dataset,
                keys_group,
                values_group,
                added_values,
                output_name_to_idx=mapping_to_idx,
            )

    def _create_hdf_input_output(
        self,
        index_dataset,  # type: int
        design_vars_group,  # type: h5py.Group
        keys_group,  # type: h5py.Group
        values_group,  # type: h5py.Group
        input_values,  # type: HashableNdarray
        output_values,  # type: Mapping[str, Union[float, ndarray, List[int]]]
    ):
        """Create the new hdf5 datasets for the given inputs and outputs.

        Useful when exporting the database to an hdf5 file.

        Args:
            index_dataset: The index of the new hdf5 entry.
            design_vars_group: The hdf5 group of the design variable values.
            keys_group: The hdf5 group of the output names.
            values_group: The hdf5 group of the output values.
            input_values: The input values.
            output_values: The output values.
        """
        self._add_hdf_input_dataset(index_dataset, design_vars_group, input_values)
        self._add_hdf_output_dataset(
            index_dataset, keys_group, values_group, output_values
        )

    def export_hdf(
        self,
        file_path="optimization_history.h5",  # type: str
        append=False,  # type: bool
    ):  # type: (...) -> None
        """Export the optimization database to a hdf file.

        Args:
            file_path: The name of the hdf file.
            append: If True, append the data to the file; False otherwise.
        """
        mode = "a" if append else "w"

        with h5py.File(file_path, mode) as h5file:
            design_vars_grp = h5file.require_group("x")
            keys_group = h5file.require_group("k")
            values_group = h5file.require_group("v")
            index_dataset = 0

            # The append mode loops over the last stored entries in order to
            # check whether some new outputs have been added.
            # However, if the hdf file has been re-written by a previous function
            # (such as OptimizationProblem.export_hdf),
            # there is no existing database inside the hdf file.
            # In such case, we have to check whether the design
            # variables group exists because otherwise the function tries to
            # append something empty.
            if append and len(design_vars_grp) != 0:
                input_values_to_idx = dict(zip(self.keys(), range(len(self.keys()))))

                for input_values in self.__hdf_export_buffer:
                    output_values = self[input_values]
                    index_dataset = input_values_to_idx[input_values]

                    if str(index_dataset) in design_vars_grp:
                        self._append_hdf_output(
                            index_dataset, keys_group, values_group, output_values
                        )
                    else:
                        self._create_hdf_input_output(
                            index_dataset,
                            design_vars_grp,
                            keys_group,
                            values_group,
                            input_values,
                            output_values,
                        )
            else:
                for input_values, output_values in self.items():
                    self._create_hdf_input_output(
                        index_dataset,
                        design_vars_grp,
                        keys_group,
                        values_group,
                        input_values,
                        output_values,
                    )
                    index_dataset += 1

        self.__hdf_export_buffer = []

    def import_hdf(
        self, filename="optimization_history.h5"  # type: str
    ):  # type: (...) -> None
        """Import a database from a hdf file.

        Args:
            filename: The path to the HDF5 file.
        """
        with h5py.File(filename, "r") as h5file:
            design_vars_grp = h5file["x"]
            keys_group = h5file["k"]
            values_group = h5file["v"]

            for raw_index in range(len(design_vars_grp)):
                str_index = str(raw_index)
                keys = [k.decode() for k in get_hdf5_group(keys_group, str_index)]

                array_name = "arr_{}".format(str_index)

                if array_name in values_group:
                    argrp = values_group[array_name]
                    vec_dict = {keys[int(k)]: array(v) for k, v in argrp.items()}
                else:
                    vec_dict = {}

                scalar_keys = (k for k in keys if k not in vec_dict)

                if str_index in values_group:
                    locvalues_scalars = get_hdf5_group(values_group, str_index)
                    scalar_dict = dict(zip(scalar_keys, locvalues_scalars))
                else:
                    scalar_dict = {}

                scalar_dict.update(vec_dict)

                self.store(
                    array(design_vars_grp[str_index]), scalar_dict, add_iter=False
                )

    @staticmethod
    def set_dv_names(
        n_dv,  # type: int
    ):  # type: (...) -> List[str]
        """Return the default input variables names.

        Args:
            n_dv: The number of variables.

        Returns:
            The names of the variables.
        """
        return ["x_" + str(i) for i in range(1, n_dv + 1)]

    def _format_design_variables_names(
        self,
        design_variables_names,  # type: Optional[str, Iterable[str]]
        dimension,  # type: int
    ):  # type: (...) -> Union[List[str], Tuple[str]]
        """Format the design variables names to be displayed in the history.

        Args:
            design_variables_names: The names of design variables.
            dimension: The dimension for default names if ``design_variables_names``
                is ``None``.

        Returns:
            The formatted names of the design variables.

        Raises:
            TypeError: If the type of ``design_variables_names`` is finally not iterable.
        """
        if design_variables_names is None:
            design_variables_names = self.set_dv_names(dimension)
        elif isinstance(design_variables_names, string_types):
            design_variables_names = [design_variables_names]
        elif not isinstance(design_variables_names, list) and not isinstance(
            design_variables_names, tuple
        ):
            raise TypeError(
                "The argument design_variables_names must be a list or a tuple whereas "
                "a {} is provided".format(type(design_variables_names))
            )
        return design_variables_names

    def get_history_array(
        self,
        functions=None,  # type: Optional[Iterable[str]]
        design_variables_names=None,  # type: Optional[Union[str, Iterable[str]]]
        add_missing_tag=False,  # type: bool
        missing_tag="NA",  # type: str
        add_dv=True,  # type: bool
        all_iterations=False,  # type: bool
        stacked_data=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> Tuple[ndarray, List[str], Iterable[str]]
        """Return the history of the optimization process.

        Args:
            functions: The names of output functions that must be returned.
            design_variables_names: The names of the input design variables.
            add_missing_tag: If True, add the tag specified in ``missing_tag`` for
                data that are not available.
            missing_tag: The tag that is added for data that are not available.
            add_dv: If True, the input design variables are returned in the history.
            all_iterations: If True, the points which are called at several
                iterations will be duplicated in the history (each duplicate
                corresponding to a different calling index); otherwise each point
                will appear only once (with the latest calling index).
            stacked_data: The names of outputs corresponding to data stored as
                ``iterable`` (e.g. iterations, penalization parameters or
                trust region radii).

        Returns:
            The values of the history.
            The names of the columns corresponding to the values of the history.
            The names of the output functions.
        """
        if functions is None:
            functions = self.get_all_data_names()
        f_history, x_history = self.get_complete_history(
            functions, add_missing_tag, missing_tag, all_iterations, stacked_data
        )
        design_variables_names = self._format_design_variables_names(
            design_variables_names, len(x_history[0])
        )
        flat_vals = []
        fdict = OrderedDict()
        for f_val_i in f_history:
            flat_vals_i = []
            for f_val, f_name in zip(f_val_i, functions):
                if isinstance(f_val, list):
                    f_val = array(f_val)
                if isinstance(f_val, ndarray) and len(f_val) > 1:
                    flat_vals_i = flat_vals_i + f_val.tolist()
                    fdict[f_name] = [
                        f_name + "_" + str(i + 1) for i in range(len(f_val))
                    ]
                else:
                    flat_vals_i.append(f_val)
                    if f_name not in fdict:
                        fdict[f_name] = [f_name]
            flat_vals.append(flat_vals_i)
        flat_names = sorted(list(chain(*fdict.values())))

        x_flat_vals = []
        xdict = OrderedDict()
        for x_val_i in x_history:
            x_flat_vals_i = []
            for x_val, x_name in zip(x_val_i, design_variables_names):
                if isinstance(x_val, ndarray) and len(x_val) > 1:
                    x_flat_vals_i = x_flat_vals_i + x_val.tolist()
                    xdict[x_name] = [
                        x_name + "_" + str(i + 1) for i in range(len(x_val))
                    ]
                else:
                    x_flat_vals_i.append(x_val)
                    if x_name not in xdict:
                        xdict[x_name] = [x_name]
            x_flat_vals.append(x_flat_vals_i)

        x_flat_names = list(chain(*xdict.values()))
        if add_dv:
            variables_names = flat_names + x_flat_names
        else:
            variables_names = flat_names

        f_history = array(flat_vals).real
        x_history = array(x_flat_vals).real
        if add_dv:
            f2d = atleast_2d(f_history)
            x2d = atleast_2d(x_history)
            if f2d.shape[0] == 1:
                f2d = f2d.T
            if x2d.shape[0] == 1:
                x2d = x2d.T
            values_array = concatenate((f2d, x2d), axis=1)
        else:
            values_array = f_history
        return values_array, variables_names, functions

    def export_to_ggobi(
        self,
        functions=None,  # type: Optional[Iterable[str]]
        file_path="opt_hist.xml",  # type: str
        design_variables_names=None,  # type: Optional[Union[str, Iterable[str]]]
    ):  # type: (...) -> None
        """Export the database to a xml file for ggobi tool.

        Args:
            functions: The names of output functions.
            file_path: The path to the xml file.
            design_variables_names: The names of the input design variables.
        """
        values_array, variables_names, functions = self.get_history_array(
            functions, design_variables_names, add_missing_tag=True, missing_tag="NA"
        )
        LOGGER.info("Export to ggobi for functions: %s", str(functions))
        LOGGER.info("Export to ggobi file: %s", str(file_path))
        save_data_arrays_to_xml(
            variables_names=variables_names,
            values_array=values_array,
            file_path=file_path,
        )

    def import_from_opendace(
        self, database_file  # type: str
    ):  # type: (...) -> None
        """Load the current database from an opendace xml database.

        Args:
            database_file: The path to an opendace database.
        """
        tree = parse_element(database_file)
        for link in tree.getroot().iter("link"):
            data = {}
            for information in link:
                for x_ydyddy in information:
                    data[x_ydyddy.tag] = literal_eval(x_ydyddy.text)
            x_vect = array(data.pop("x"))
            data_reformat = data["y"]
            for key, value in data["dy"].items():
                data_reformat["@" + key[1:]] = array(value)
            self.store(x_vect, data_reformat)

    @classmethod
    def get_gradient_name(
        cls,
        name,  # type: str
    ):  # type: (...) -> str
        """Return the name of the gradient related to a function.

        This name is the concatenation of a GRAD_TAG, e.g. '@',
        and the name of the function, e.g. 'f'.
        With this example, the name of the gradient is '@f'.

        Args:
            name: The name of a function.

        Returns:
            The name of the gradient based on the name of the function.
        """
        return "{}{}".format(cls.GRAD_TAG, name)

    def __str__(self):  # type: (...) -> str
        """Return the string representation.

        The string representation of the database is based on the underlying dictionary
        string representation.
        """
        return str(self.__dict)


class HashableNdarray(object):
    """HashableNdarray wrapper for ndarray objects.

    Instances of ndarray are not HashableNdarray,
    meaning they cannot be added to sets,
    nor used as keys in dictionaries.
    This is by design, ndarray objects are mutable,
    and therefore cannot reliably implement the __hash__() method.

    The HashableNdarray class allows a way around this limitation.
    It implements the required methods for HashableNdarray
    objects in terms of an encapsulated ndarray object.
    This can be either a copied instance (which is safer)
    or the original object
    (which requires the user to be careful enough not to modify it).
    """

    def __init__(
        self,
        wrapped,  # type: ndarray
        tight=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            wrapped: The array that must be wrapped.
            tight: If True, the wrapped array is copied.
        """
        self.__tight = tight
        self.wrapped = array(wrapped) if tight else wrapped
        self.__hash = int(sha1(wrapped.view(uint8)).hexdigest(), 16)

    def __eq__(
        self, other  # type: HashableNdarray
    ):  # type: (...) -> bool
        """Check equality with another HashableNdarray.

        Args:
            other: The other hashable array.

        Returns:
            True if the two arrays are equal, False otherwise.
        """
        return all(self.wrapped == other.wrapped)

    def __hash__(self):  # type: (...) -> int
        """Return the hash number of the current array.

        Returns:
            The hash number.
        """
        return self.__hash

    def unwrap(self):  # type: (...) -> ndarray
        """Return the encapsulated ndarray.

        Returns:
            The encapsulated ndarray, or a copy if the wrapper is ``tight``.
        """
        if self.__tight:
            return array(self.wrapped)

        return self.wrapped

    def __str__(self):  # type: (...) -> str
        """Return the informal string representation of the array.

        Returns:
            The string representation.
        """
        return str(array(self.wrapped))

    def __repr__(self):  # type: (...) -> str
        """Return the official string representation.

        Returns:
            The string representation.
        """
        return str(self)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Gradient approximation."""
from __future__ import division, unicode_literals

import logging
from typing import Any, Callable, List, Optional, Sequence, Tuple, Union

import six
from custom_inherit import DocInheritMeta
from numpy import array, finfo, float64, ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.core.factory import Factory
from gemseo.utils.py23_compat import xrange

EPSILON = finfo(float).eps
LOGGER = logging.getLogger(__name__)


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class GradientApproximator(object):
    """A gradient approximator.

    Attributes:
        f_pointer (Callable): The pointer to the function to derive.
    """

    ALIAS = None

    def __init__(
        self,
        f_pointer,  # type: Callable[[ndarray],ndarray]
        step=1e-6,  # type: float
        parallel=False,  # type: bool
        design_space=None,  # type: Optional[DesignSpace]
        normalize=True,  # type: bool
        **parallel_args  # type: Union[int,bool,float]
    ):  # type: (...) -> None
        """
        Args:
            f_pointer: The pointer to the function to derive.
            step: The default differentiation step.
            parallel: Whether to differentiate the function in parallel.
            design_space: The design space
                containing the upper bounds of the input variables.
                If None, consider that the input variables are unbounded.
            normalize: If True, then the functions are normalized.
            **parallel_args: The parallel execution options,
                see :mod:`gemseo.core.parallel_execution`.
        """
        self.f_pointer = f_pointer
        self.__par_args = parallel_args
        self.__parallel = parallel
        self._step = None
        self.step = step
        self._design_space = design_space
        self._normalize = normalize

    @property
    def _parallel(self):  # type: (...) -> bool
        """Whether to differentiate the function in parallel."""
        return self.__parallel

    @property
    def _par_args(self):  # type: (...) -> Union[int,bool,float]
        """The parallel execution options."""
        return self.__par_args

    @property
    def step(self):  # type: (...) -> float
        """The default approximation step."""
        return self._step

    @step.setter
    def step(
        self,
        value,  # type: float
    ):  # type: (...) -> None
        self._step = value

    def f_gradient(
        self,
        x_vect,  # type: ndarray
        step=None,  # type: Optional[float]
        x_indices=None,  # type: Optional[Sequence[int]]
        **kwargs  # type: Any
    ):  # type: (...) -> ndarray
        """Approximate the gradient of the function for a given input vector.

        Args:
            x_vect: The input vector.
            step: The differentiation step.
                If None, use the default differentiation step.
            x_indices: The components of the input vector
                to be used for the differentiation.
                If None, use all the components.
            **kwargs: The optional arguments for the function.

        Returns:
            The approximated gradient.
        """
        input_dimension = len(x_vect)
        input_perturbations, steps = self.generate_perturbations(
            input_dimension, x_vect, x_indices=x_indices, step=step
        )
        n_perturbations = input_perturbations.shape[1]

        if self._parallel:
            grad = self._compute_parallel_grad(
                x_vect, n_perturbations, input_perturbations, steps, **kwargs
            )
        else:
            grad = self._compute_grad(
                x_vect, n_perturbations, input_perturbations, steps, **kwargs
            )

        return array(grad, dtype=float64).T

    def _compute_parallel_grad(
        self,
        input_values,  # type: ndarray
        n_perturbations,  # type: int
        input_perturbations,  # type: ndarray
        step,  # type: float
        **kwargs  # type: Any
    ):  # type: (...) -> ndarray
        """Approximate the gradient in parallel.

        Args:
            input_values: The input values.
            n_perturbations: The number of perturbations.
            step: The differentiation step,
                either one global step or one step by input component.
            **kwargs: The optional arguments for the function.

        Returns:
            The approximated gradient.
        """
        raise NotImplementedError

    def _compute_grad(
        self,
        input_values,  # type: ndarray
        n_perturbations,  # type: int
        input_perturbations,  # type: ndarray
        step,  # type: Union[float,ndarray]
        **kwargs  # type: Any
    ):  # type: (...) -> ndarray
        """Approximate the gradient.

        Args:
            input_values: The input values.
            n_perturbations: The number of perturbations.
            input_perturbations: The input perturbations.
            step: The differentiation step,
                either one global step or one step by input component.
            **kwargs: The optional arguments for the function.

        Returns:
            The approximated gradient.
        """
        raise NotImplementedError

    def generate_perturbations(
        self,
        n_dim,  # type: int
        x_vect,  # type: ndarray
        x_indices=None,  # type: Optional[Sequence[int]]
        step=None,  # type: Optional[float]
    ):  # type: (...) -> Tuple[ndarray,Union[float,ndarray]]
        """Generate the input perturbations from the differentiation step.

        These perturbations will be used to compute the output ones.

        Args:
            n_dim: The input dimension.
            x_vect: The input vector.
            x_indices: The components of the input vector
                to be used for the differentiation.
                If None, use all the components.
            step: The differentiation step.
                If None, use the default differentiation step.

        Returns:
            * The input perturbations.
            * The differentiation step,
              either one global step or one step by input component.
        """
        if step is None:
            step = self.step

        if x_indices is None:
            x_indices = xrange(n_dim)

        return self._generate_perturbations(x_vect, x_indices, step)

    def _generate_perturbations(
        self,
        input_values,  # type: ndarray
        input_indices,  # type: List[int]
        step,  # type: float
    ):  # type: (...) -> Tuple[ndarray,Union[float,ndarray]]
        """Generate the input perturbations from the differentiation step.

        These perturbations will be used to compute the output ones.

        Args:
            input_values: The input vector.
            input_indices: The components of the input vector
                to be used for the differentiation.
            step: The differentiation step.

        Returns:
            * The input perturbations.
            * The differentiation step,
              either one global step or one step by input component.
        """
        raise NotImplementedError


class GradientApproximationFactory(object):
    """A factory to create gradient approximators."""

    def __init__(self):  # type: (...) -> None
        self.factory = Factory(GradientApproximator, ("gemseo.utils.derivatives",))
        self.__aliases = {
            self.factory.get_class(class_name).ALIAS: class_name
            for class_name in self.factory.classes
        }

    def create(
        self,
        name,  # type:str
        f_pointer,  # type: Callable
        step=None,  # type: Optional[float]
        parallel=False,  # type: bool
        **parallel_args
    ):  # type: (...) -> GradientApproximator
        """Create a gradient approximator.

        Args:
            name: Either the name of the class implementing the gradient approximator
                or its alias.
            f_pointer: The pointer to the function to differentiate.
            step: The differentiation step.
                If None, use a default differentiation step.
            parallel: Whether to differentiate the function in parallel.
            **parallel_args: The parallel execution options,
                see :mod:`gemseo.core.parallel_execution`.

        Returns:
            The gradient approximator.
        """
        if name in self.__aliases:
            name = self.__aliases[name]

        if step is None:
            return self.factory.create(
                name, f_pointer=f_pointer, parallel=parallel, **parallel_args
            )
        else:
            return self.factory.create(
                name, f_pointer=f_pointer, step=step, parallel=parallel, **parallel_args
            )

    @property
    def gradient_approximators(self):  # type: (...) -> List[str]
        """The gradient approximators."""
        return self.factory.classes

    def is_available(self, class_name):  # type:(...) -> bool
        """Whether a gradient approximator is available.

        Args:
            class_name: The name of the class implementing the gradient approximator.

        Return:
            Whether the gradient approximator is available.
        """
        return self.factory.is_available(class_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Introduction
************

Here is the Application Programming Interface (API) of |g|,
a set of high level functions for ease of use.

Make |g| ever more accessible
-----------------------------

The aim of this API is to provide high level functions that are sufficient
to use |g| in most cases, without requiring a deep knowledge of |g|.

Besides, these functions shall change much less often than the internal
classes, which is key for backward compatibility,
which means ensuring that your current scripts using |g| will be usable
with the future versions of |g|.

Connect |g| to your favorite tools
----------------------------------

The API also facilitates the interfacing of |g|
with a platform or other software.

To interface a simulation software with |g|,
please refer to: :ref:`software_connection`.

Extending |g|
-------------

See :ref:`extending-gemseo`.

Table of contents
*****************

Algorithms
----------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.execute_algo`
- :meth:`~gemseo.api.get_available_opt_algorithms`
- :meth:`~gemseo.api.get_available_doe_algorithms`
- :meth:`~gemseo.api.get_algorithm_options_schema`

Cache
-----

- :meth:`~gemseo.api.create_cache`
- :meth:`~gemseo.api.get_available_caches`

Configuration
-------------

- :meth:`~gemseo.api.configure_logger`

Coupling
--------

- :meth:`~gemseo.api.generate_n2_plot`
- :meth:`~gemseo.api.generate_coupling_graph`
- :meth:`~gemseo.api.get_all_inputs`
- :meth:`~gemseo.api.get_all_outputs`

Design space
------------

- :meth:`~gemseo.api.read_design_space`
- :meth:`~gemseo.api.export_design_space`
- :meth:`~gemseo.api.create_design_space`

Disciplines
-----------

- :meth:`~gemseo.api.create_discipline`
- :meth:`~gemseo.api.import_discipline`
- :meth:`~gemseo.api.get_available_disciplines`
- :meth:`~gemseo.api.get_discipline_inputs_schema`
- :meth:`~gemseo.api.get_discipline_outputs_schema`
- :meth:`~gemseo.api.get_discipline_options_schema`
- :meth:`~gemseo.api.get_discipline_options_defaults`

Formulations
------------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.get_available_formulations`
- :meth:`~gemseo.api.get_formulation_options_schema`
- :meth:`~gemseo.api.get_formulation_sub_options_schema`
- :meth:`~gemseo.api.get_formulations_sub_options_defaults`
- :meth:`~gemseo.api.get_formulations_options_defaults`

MDA
---

- :meth:`~gemseo.api.create_mda`
- :meth:`~gemseo.api.get_available_mdas`
- :meth:`~gemseo.api.get_mda_options_schema`

Post-processing
---------------

- :meth:`~gemseo.api.execute_post`
- :meth:`~gemseo.api.get_available_post_processings`
- :meth:`~gemseo.api.get_post_processing_options_schema`

Scalable
--------

- :meth:`~gemseo.api.create_scalable`

Scenario
--------

- :meth:`~gemseo.api.create_scenario`
- :meth:`~gemseo.api.monitor_scenario`
- :meth:`~gemseo.api.get_available_scenario_types`
- :meth:`~gemseo.api.get_scenario_options_schema`
- :meth:`~gemseo.api.get_scenario_inputs_schema`
- :meth:`~gemseo.api.get_scenario_differentiation_modes`

Surrogates
----------

- :meth:`~gemseo.api.create_surrogate`
- :meth:`~gemseo.api.get_available_surrogates`
- :meth:`~gemseo.api.get_surrogate_options_schema`

API functions
*************
"""
from __future__ import division, unicode_literals

import logging
import re
from typing import (  # noqa F401
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from numpy import ndarray
from six import string_types

from gemseo.algos.doe.doe_factory import DOEFactory
from gemseo.algos.doe.doe_lib import DOELibraryOptionType

if TYPE_CHECKING:
    from logging import Logger
    from matplotlib.figure import Figure
    from gemseo.algos.design_space import DesignSpace
    from gemseo.algos.opt_problem import OptimizationProblem
    from gemseo.algos.opt_result import OptimizationResult
    from gemseo.algos.parameter_space import ParameterSpace  # noqa:F401
    from gemseo.core.cache import AbstractCache
    from gemseo.core.dataset import Dataset
    from gemseo.core.discipline import MDODiscipline
    from gemseo.core.grammars.json_grammar import JSONGrammar
    from gemseo.core.scenario import Scenario
    from gemseo.core.surrogate_disc import SurrogateDiscipline  # noqa:F401
    from gemseo.mda.mda import MDA
    from gemseo.mlearning.core.ml_algo import TransformerType
    from gemseo.problems.scalable.data_driven.discipline import (  # noqa:F401
        ScalableDiscipline,
    )

from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.logging_tools import MultiLineFileHandler, MultiLineStreamHandler
from gemseo.utils.py23_compat import Path

# Most modules are imported directly in the methods, which adds a very small
# overhead, but prevents users from importing them from the API.
# All factories are Singletons which means that the scan for
# plugins is done once only

LOGGER = logging.getLogger(__name__)

# pylint: disable=import-outside-toplevel


def generate_n2_plot(
    disciplines,  # type: Sequence[MDODiscipline]
    file_path="n2.pdf",  # type: Union[str,Path]
    show_data_names=True,  # type: bool
    save=True,  # type: bool
    show=False,  # type: bool
    figsize=(15, 10),  # type: Tuple[int]
    open_browser=False,  # type: bool
):  # type: (...) -> None
    """Generate a N2 plot from disciplines.

    Args:
        disciplines: The disciplines from which the N2 chart is generated.
        file_path: The path of the file to save the figure.
        show_data_names: If ``True``, show the names of the coupling data ;
            otherwise,
            circles are drawn,
            the size of which depends on the number of coupling names.
        save: If True, save the figure to the file_path.
        show: If True, show the plot.
        figsize: The width and height of the figure.
        open_browser: If True, open a browser and display an interactive N2 chart.

    Examples
    --------
    >>> from gemseo.api import create_discipline, generate_n2_plot
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> generate_n2_plot(disciplines)

    See also
    --------
    generate_coupling_graph
    get_all_inputs
    get_all_outputs
    """
    from gemseo.core.coupling_structure import MDOCouplingStructure

    coupling_structure = MDOCouplingStructure(disciplines)
    coupling_structure.plot_n2_chart(
        file_path, show_data_names, save, show, figsize, open_browser
    )


def generate_coupling_graph(
    disciplines,  # type: Sequence[MDODiscipline]
    file_path="coupling_graph.pdf",  # type: Union[str,Path]
    full=True,  # type: bool
):  # type: (...) -> None
    """Generate a graph of the couplings between disciplines.

    Args:
        disciplines: The disciplines from which the graph is generated.
        file_path: The path of the file to save the figure.
        full: If True, generate the full coupling graph.
            Otherwise, generate the condensed one.

    Examples
    --------
    >>> from gemseo.api import create_discipline, generate_coupling_graph
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> generate_coupling_graph(disciplines)

    See also
    --------
    generate_n2_plot
    get_all_inputs
    get_all_outputs
    """
    from gemseo.core.coupling_structure import MDOCouplingStructure

    coupling_structure = MDOCouplingStructure(disciplines)
    if full:
        coupling_structure.graph.export_initial_graph(file_path)
    else:
        coupling_structure.graph.export_reduced_graph(file_path)


def get_available_formulations():  # type: (...) -> List[str]
    """Return the names of the available formulations.

    Returns:
        The names of the available MDO formulations.

    Examples
    --------
    >>> from gemseo.api import get_available_formulations
    >>> get_available_formulations()

    See also
    --------
    create_scenario
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    return MDOFormulationsFactory().formulations


def get_available_opt_algorithms():  # type: (...) -> List[str]
    """Return the names of the available optimization algorithms.

    Returns:
        The names of the available optimization algorithms.

    Examples
    --------
    >>> from gemseo.api import get_available_opt_algorithms
    >>> get_available_opt_algorithms()

    See also
    --------
    create_scenario
    execute_algo
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.opt.opt_factory import OptimizersFactory

    return OptimizersFactory().algorithms


def get_available_doe_algorithms():  # type: (...) -> List[str]
    """Return the names of the available design of experiments (DOEs) algorithms.

    Returns:
        The names of the available DOE algorithms.

    Examples
    --------
    >>> from gemseo.api import get_available_doe_algorithms
    >>> get_available_doe_algorithms()

    See also
    --------
    create_scenario
    execute_algo
    get_available_opt_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.doe.doe_factory import DOEFactory

    return DOEFactory().algorithms


def get_available_surrogates():  # type: (...) -> List[str]
    """Return the names of the available surrogate disciplines.

    Returns:
        The names of the available surrogate disciplines.

    Examples
    --------
    >>> from gemseo.api import get_available_surrogates
    >>> print(get_available_surrogates())
    ['RBFRegression', 'GaussianProcessRegression', 'LinearRegression', 'PCERegression']

    See also
    --------
    create_surrogate
    get_surrogate_options_schema
    """
    from gemseo.mlearning.api import get_regression_models

    return get_regression_models()


def get_available_disciplines():  # type: (...) -> List[str]
    """Return the names of the available disciplines.

    Returns:
        The names of the available disciplines.

    Examples
    --------
    >>> from gemseo.api import get_available_disciplines
    >>> print(get_available_disciplines())
    ['RosenMF', 'SobieskiAerodynamics', 'ScalableKriging', 'DOEScenario',
    'MDOScenario', 'SobieskiMission', 'SobieskiBaseWrapper', 'Sellar1',
    'Sellar2', 'MDOChain', 'SobieskiStructure', 'AutoPyDiscipline',
    'Structure', 'SobieskiPropulsion', 'Scenario', 'AnalyticDiscipline',
    'MDOScenarioAdapter', 'ScalableDiscipline', 'SellarSystem', 'Aerodynamics',
    'Mission', 'PropaneComb1', 'PropaneComb2', 'PropaneComb3',
    'PropaneReaction', 'MDOParallelChain']

    See also
    --------
    create_discipline
    import_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    return DisciplinesFactory().disciplines


def get_surrogate_options_schema(
    surrogate_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the available options for a surrogate discipline.

    Args:
        surrogate_name: The name of the surrogate discipline.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the surrogate discipline.

    Examples
    --------
    >>> from gemseo.api import get_surrogate_options_schema
    >>> tmp = get_surrogate_options_schema('LinRegSurrogateDiscipline',
    >>>                                    pretty_print=True)

    See also
    --------
    create_surrogate
    get_available_surrogates
    """
    from gemseo.mlearning.api import get_regression_options

    return get_regression_options(surrogate_name, output_json, pretty_print)


def get_algorithm_options_schema(
    algorithm_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the options of an algorithm.

    Args:
        algorithm_name: The name of the algorithm.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the algorithm.

    Raises:
        ValueError: When the algorithm is not available.

    Examples
    --------
    >>> from gemseo.api import get_algorithm_options_schema
    >>> schema = get_algorithm_options_schema('NLOPT_SLSQP', pretty_print=True)

    See also
    --------
    create_scenario
    execute_algo
    get_available_opt_algorithms
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    from gemseo.algos.doe.doe_factory import DOEFactory
    from gemseo.algos.opt.opt_factory import OptimizersFactory

    for factory in (DOEFactory(), OptimizersFactory()):
        if factory.is_available(algorithm_name):
            algo_lib = factory.create(algorithm_name)
            opts_gram = algo_lib.init_options_grammar(algorithm_name)
            schema = _get_schema(opts_gram, output_json, pretty_print)
            return schema
    raise ValueError("Algorithm named {} is not available.".format(algorithm_name))


def get_discipline_inputs_schema(
    discipline,  # type: MDODiscipline
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the inputs of a discipline.

    Args:
        discipline: The discipline.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the inputs of the discipline.

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_discipline_inputs_schema
    >>> discipline = create_discipline('Sellar1')
    >>> schema = get_discipline_inputs_schema(discipline, pretty_print=True)

    See also
    --------
    create_discipline
    import_discipline
    get_available_disciplines
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    return _get_schema(discipline.input_grammar, output_json, pretty_print)


def get_discipline_outputs_schema(
    discipline,  # type: MDODiscipline
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the outputs of a discipline.

    Args:
        discipline: The discipline.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the outputs of the discipline.

    Examples
    --------
    >>> from gemseo.api import get_discipline_outputs_schema, create_discipline
    >>> discipline = create_discipline('Sellar1')
    >>> get_discipline_outputs_schema(discipline, pretty_print=True)

    See also
    --------
    create_discipline
    import_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    return _get_schema(discipline.output_grammar, output_json, pretty_print)


def get_available_post_processings():  # type: (...) -> List[str]
    """Return the names of the available optimization post-processings.

    Returns:
        The names of the available post-processings.

    Examples
    --------
    >>> from gemseo.api import get_available_post_processings
    >>> print(get_available_post_processings())
    ['ScatterPlotMatrix', 'VariableInfluence', 'ConstraintsHistory',
    'RadarChart', 'Robustness', 'Correlations', 'SOM', 'KMeans',
    'ParallelCoordinates', 'GradientSensitivity', 'OptHistoryView',
    'BasicHistory', 'ObjConstrHist', 'QuadApprox']

    See also
    --------
    execute_post
    get_post_processing_options_schema
    """
    from gemseo.post.post_factory import PostFactory

    return PostFactory().posts


def get_post_processing_options_schema(
    post_proc_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the options of a post-processing.

    Args:
        post_proc_name: The name of the post-processing.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the post-processing.

    Examples
    --------
    >>> from gemseo.api import get_post_processing_options_schema
    >>> schema = get_post_processing_options_schema('OptHistoryView',
    >>>                                             pretty_print=True)

    See also
    --------
    execute_post
    get_available_post_processings
    """
    from gemseo.algos.opt_problem import OptimizationProblem
    from gemseo.post.post_factory import PostFactory

    post_proc = PostFactory().create(OptimizationProblem(None), post_proc_name)
    return _get_schema(post_proc.opt_grammar, output_json, pretty_print)


def get_formulation_options_schema(
    formulation_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the options of a formulation.

    Args:
        formulation_name: The name of the formulation.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulation_options_schema
    >>> schema = get_formulation_options_schema('MDF', pretty_print=True)

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    grammar = factory.get_options_grammar(formulation_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_formulation_sub_options_schema(
    formulation_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
    **formulation_options  # type: Any
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the sub-options of a formulation.

    Args:
        formulation_name: The name of the formulation.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.
        **formulation_options: The options of the formulation
            required for its instantiation.

    Returns:
        The schema of the sub-options of the formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulation_sub_options_schema
    >>> schema = get_formulation_sub_options_schema('MDF',
    >>>                                             main_mda_class='MDAJacobi',
    >>>                                             pretty_print=True)

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulations_options_defaults
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    grammar = factory.get_sub_options_grammar(formulation_name, **formulation_options)
    return _get_schema(grammar, output_json, pretty_print)


def get_formulations_sub_options_defaults(
    formulation_name,  # type: str
    **formulation_options  # type: Any
):  # type: (...) -> Dict[str,Any]
    """Return the default values of the sub-options of a formulation.

    Args:
        formulation_name: The name of the formulation.
        **formulation_options: The options of the formulation
            required for its instantiation.

    Returns:
        The default values of the sub-options of the formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulations_sub_options_defaults
    >>> get_formulations_sub_options_defaults('MDF',
    >>>                                       main_mda_class='MDAJacobi')

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    return factory.get_default_sub_options_values(
        formulation_name, **formulation_options
    )


def get_formulations_options_defaults(
    formulation_name,  # type: str
):  # type: (...) -> Dict[str,Any]
    """Return the default values of the options of a formulation.

    Args:
        formulation_name: The name of the formulation.
        **formulation_options: The options of the formulation
            required for its instantiation.

    Returns:
        The default values of the options of the formulation.

    Examples
    --------
    >>> from gemseo.api import get_formulations_options_defaults
    >>> get_formulations_options_defaults('MDF')
    {'main_mda_class': 'MDAChain',
     'maximize_objective': False,
     'sub_mda_class': 'MDAJacobi'}

    See also
    --------
    create_scenario
    get_available_formulations
    get_formulation_options_schema
    get_formulation_sub_options_schema
    get_formulations_sub_options_defaults
    """
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory

    factory = MDOFormulationsFactory().factory
    return factory.get_default_options_values(formulation_name)


def get_discipline_options_schema(
    discipline_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of a discipline.

    Args:
        discipline_name: The name of the formulation.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the discipline.

    Examples
    --------
    >>> from gemseo.api import get_discipline_options_schema
    >>> schema = get_discipline_options_schema('Sellar1', pretty_print=True)

    See also
    --------
    create_discipline
    import_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    disc_fact = DisciplinesFactory()
    grammar = disc_fact.get_options_grammar(discipline_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_scenario_options_schema(
    scenario_type,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the options of a scenario.

    Args:
        scenario_type: The type of the scenario.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the scenario.

    Examples
    --------
    >>> from gemseo.api import get_scenario_options_schema
    >>> get_scenario_options_schema('MDO')

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differentiation_modes
    """
    if scenario_type not in get_available_scenario_types():
        raise ValueError("Unknown scenario type {}".format(scenario_type))
    scenario_class = {"MDO": "MDOScenario", "DOE": "DOEScenario"}[scenario_type]
    return get_discipline_options_schema(scenario_class, output_json, pretty_print)


def get_scenario_inputs_schema(
    scenario,  # type: Scenario
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the inputs of a scenario.

    Args:
        scenario: The scenario.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the inputs of the scenario.

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario,
    get_scenario_inputs_schema
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> design_space = SellarDesignSpace()
    >>> disciplines = create_discipline(['Sellar1','Sellar2','SellarSystem'])
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    'my_scenario', 'MDO')
    >>> get_scenario_inputs_schema(scenario)

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_differentiation_modes
    """
    return get_discipline_inputs_schema(scenario, output_json, pretty_print)


def get_discipline_options_defaults(
    discipline_name,  # type: str
):  # type: (...) -> Dict[str,Any]
    """Return the default values of the options of a discipline.

    Args:
        discipline_name: The name of the discipline.

    Returns:
        The default values of the options of the discipline.

    Examples
    --------
    >>> from gemseo.api import get_discipline_options_defaults
    >>> get_discipline_options_defaults('Sellar1')

    See also
    --------
    create_discipline
    import_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    factory = DisciplinesFactory().factory
    return factory.get_default_options_values(discipline_name)


def get_scenario_differentiation_modes():
    """Return the names of the available differentiation modes of a scenario.

    Returns:
        The names of the available differentiation modes of a scenario.

    Examples
    --------
    >>> from gemseo.api import get_scenario_differentiation_modes
    >>> get_scenario_differentiation_modes()

    See also
    --------
    create_scenario
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    """
    from gemseo.algos.opt_problem import OptimizationProblem

    return OptimizationProblem.DIFFERENTIATION_METHODS


# TODO: to be deprecated
get_scenario_differenciation_modes = get_scenario_differentiation_modes


def get_available_scenario_types():  # type: (...) -> List[str]
    """Return the names of the available scenario types.

    Returns:
        The names of the available scenario types.

    Examples
    --------
    >>> from gemseo.api import get_available_scenario_types
    >>> get_available_scenario_types()

    See also
    --------
    create_scenario
    monitor_scenario
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differentiation_modes
    """
    return ["MDO", "DOE"]


def _get_schema(
    json_grammar,  # type: Optional[JSONGrammar]
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Optional[Union[str,Dict[str,Any]]]
    """Return the schema of a JSON grammar.

    Args:
        json_grammar: The JSON grammar to be considered.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the JSON grammar if any.
    """
    if json_grammar is None:
        return None
    schema = json_grammar.schema
    dict_schema = schema.to_dict()
    if pretty_print:
        if "name" in dict_schema:
            title = dict_schema["name"].replace("_", " ")
        else:
            title = None
        table = PrettyTable(title=title, max_table_width=150)
        names = []
        descriptions = []
        types = []
        for name, value in dict_schema["properties"].items():
            names.append(name)
            descriptions.append(value.get("description"))
            description = descriptions[-1]
            tmp = []
            if descriptions[-1] is not None:
                descriptions[-1] = descriptions[-1].split(":type")[0]
                descriptions[-1] = descriptions[-1].capitalize()
                descriptions[-1] = descriptions[-1].replace("\n", " ")
                tmp = re.split(r":type ([*\w]+): (.*?)", description)
            if len(tmp) == 4:
                types.append(tmp[3].strip())
            else:
                types.append(value.get("type"))
        table.add_column("Name", names)
        table.add_column("Description", descriptions)
        table.add_column("Type", types)
        table.sortby = "Name"
        table.min_width = 25
        print(table)  # noqa: T001
        LOGGER.info("%s", table)
    if output_json:
        return schema.to_json()
    return dict_schema


def get_available_mdas():  # type: (...) -> List[str]
    """Return the names of the available multidisciplinary analyses (MDAs).

    Returns:
        The names of the available MDAs.

    Examples
    --------
    >>> from gemseo.api import get_available_mdas
    >>> get_available_mdas()

    See also
    --------
    create_mda
    get_mda_options_schema
    """
    from gemseo.mda.mda_factory import MDAFactory

    return MDAFactory().mdas


def get_mda_options_schema(
    mda_name,  # type: str
    output_json=False,  # type: bool
    pretty_print=False,  # type: bool
):  # type: (...) -> Union[str,Dict[str,Any]]
    """Return the schema of the options of a multidisciplinary analysis (MDA).

    Args:
        mda_name: The name of the MDA.
        output_json: Whether to apply the JSON format to the schema.
        pretty_print: Whether to print the schema in a tabular way.

    Returns:
        The schema of the options of the MDA.

    Examples
    --------
    >>> from gemseo.api import get_mda_options_schema
    >>> get_mda_options_schema('MDAJacobi')

    See also
    --------
    create_mda
    get_available_mdas
    """
    from gemseo.mda.mda_factory import MDAFactory

    factory = MDAFactory().factory
    grammar = factory.get_options_grammar(mda_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_all_inputs(
    disciplines,  # type: Iterable[MDODiscipline]
    recursive=False,  # type: bool
):  # type: (...) -> List[str]
    """Return all the input names of the disciplines.

    Args:
        disciplines: The disciplines.
        recursive: If True,
            search for the inputs of the sub-disciplines,
            when some disciplines are scenarios.

    Returns:
        The names of the inputs.

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_all_inputs
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> get_all_inputs(disciplines)
    ['y_0', 'x_shared', 'y_1', 'x_local']

    See also
    --------
    generate_n2_plot
    generate_coupling_graph
    get_all_outputs
    """
    from gemseo.utils.data_conversion import DataConversion

    return DataConversion.get_all_inputs(disciplines, recursive)


def get_all_outputs(
    disciplines,  # type: Iterable[MDODiscipline]
    recursive=False,  # type: bool
):  # type: (...) -> List[str]
    """Return all the output names of the disciplines.

    Args:
        disciplines: The disciplines.
        recursive: If True,
            search for the outputs of the sub-disciplines,
            when some disciplines are scenarios.

    Returns:
        The names of the outputs.

    Examples
    --------
    >>> from gemseo.api import create_discipline, get_all_outputs
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> get_all_outputs(disciplines)
    ['y_1', 'y_0']

    See also
    --------
    generate_n2_plot
    generate_coupling_graph
    get_all_inputs
    """
    from gemseo.utils.data_conversion import DataConversion

    return DataConversion.get_all_outputs(disciplines, recursive)


def create_scenario(
    disciplines,  # type: Sequence[MDODiscipline]
    formulation,  # type: str
    objective_name,  # type: str
    design_space,  # type: Union[DesignSpace,str,Path]
    name=None,  # type: Optional[str]
    scenario_type="MDO",  # type: str
    grammar_type="JSONGrammar",  # type: str
    maximize_objective=False,  # type: bool
    **options  # type: Any
):  # type: (...) -> Scenario
    """Initialize a scenario.

    Args:
        disciplines: The disciplines
            used to compute the objective, constraints and observables
            from the design variables.
        formulation: The name of the MDO formulation,
            also the name of a class inheriting from :class:`.MDOFormulation`.
        objective_name: The name of the objective.
        design_space: The design space.
        name: The name to be given to this scenario.
            If None, use the name of the class.
        scenario_type: The type of the scenario, e.g. "MDO" or "DOE".
        grammar_type: The type of grammar to use for IO declaration,
            e.g. "JSONGrammar" or "SimpleGrammar".
        maximize_objective: Whether to maximize the objective.
        **options: The options
            to be passed to the :class:`.MDOFormulation`.

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> design_space = SellarDesignSpace()
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    >>>                            'SellarMDFScenario', 'MDO')

    See also
    --------
    monitor_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differentiation_modes
    """
    from gemseo.core.doe_scenario import DOEScenario
    from gemseo.core.mdo_scenario import MDOScenario

    if not isinstance(disciplines, list):
        disciplines = [disciplines]

    if isinstance(design_space, (string_types, Path)):
        design_space = read_design_space(design_space)

    if scenario_type == "MDO":
        return MDOScenario(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            grammar_type=grammar_type,
            maximize_objective=maximize_objective,
            **options
        )

    if scenario_type == "DOE":
        return DOEScenario(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            grammar_type=grammar_type,
            maximize_objective=maximize_objective,
            **options
        )

    raise ValueError(
        "Unknown scenario type: {}, use one of : 'MDO' or 'DOE'.".format(scenario_type)
    )


def configure_logger(
    logger_name=None,  # type: Optional[str]
    level=logging.INFO,  # type: str
    date_format="%H:%M:%S",  # type: str
    message_format="%(levelname)8s - %(asctime)s: %(message)s",  # type: str
    filename=None,  # type: Optional[Union[str,Path]]
    filemode="a",  # type: str
):  # type: (...) -> Logger
    """Configure |g| logging.

    Args:
        logger_name: The name of the logger to configure, i.e. the root logger.
        level: The logging level, either 'DEBUG', 'INFO', 'WARNING' and 'CRITICAL'.
        date_format: The logging date format.
        message_format: The logging message format.
        filename: The path to the log file, if outputs must be written in a file.
        filemode: The logging output file mode,
            either 'w' (overwrite) or 'a' (append).

    Examples
    --------
    >>> import logging
    >>> configure_logger(logging.WARNING)
    """
    if logger_name == "GEMSEO":
        # TODO: deprecate this at some point.
        # For backward compatibility, create the logger named after the modules
        # and set an alias pointing to the same logger instance.
        logger = logging.getLogger("gemseo")
        logging.Logger.manager.loggerDict["GEMSEO"] = logger
    else:
        logger = logging.getLogger(logger_name)

    logger.setLevel(level)
    formatter = logging.Formatter(fmt=message_format, datefmt=date_format)

    # remove all existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    stream_handler = MultiLineStreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    if filename is not None:
        file_handler = MultiLineFileHandler(
            filename, mode=filemode, delay=True, encoding="utf-8"
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger


def create_discipline(
    discipline_name,  # type: Union[str,Iterable[str]]
    **options  # type: Any
):
    """Instantiate one or more disciplines.

    Args:
        discipline_name: Either the name of a discipline
            or the names of several disciplines.
        **options: The options to be passed to the disciplines constructors.

    Returns:
        The disciplines.

    Examples
    --------
    >>> from gemseo.api import create_discipline
    >>> discipline = create_discipline('Sellar1')
    >>> discipline.execute()
    {'x_local': array([0.+0.j]),
     'x_shared': array([1.+0.j, 0.+0.j]),
     'y_0': array([0.89442719+0.j]),
     'y_1': array([1.+0.j])}

    See also
    --------
    import_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    factory = DisciplinesFactory()
    if isinstance(discipline_name, string_types):
        return factory.create(discipline_name, **options)

    return [factory.create(d_name, **options) for d_name in discipline_name]


def import_discipline(
    file_path,  # type: Union[str,Path]
):  # type: (...) -> MDODiscipline
    """Import a discipline from a pickle file.

    Args:
        file_path: The path to the file containing the discipline
            saved with the method :meth:`.MDODiscipline.serialize`.

    Returns:
        The discipline.

    See also
    --------
    create_discipline
    get_available_disciplines
    get_discipline_inputs_schema
    get_discipline_outputs_schema
    get_discipline_options_schema
    get_discipline_options_defaults
    """
    from gemseo.core.discipline import MDODiscipline

    return MDODiscipline.deserialize(file_path)


def create_scalable(
    name,  # type: str
    data,  # type: Dataset
    sizes=None,  # type: Mapping[str,int]
    **parameters  # type: Any
):  # type: (...) -> ScalableDiscipline
    """Create a scalable discipline from a dataset.

    Args:
        name: The name of the class of the scalable model.
        data: The learning dataset.
        sizes: The sizes of the input and output variables.
        **parameters: The parameters of the scalable model.

    Returns:
        The scalable discipline.
    """
    from gemseo.problems.scalable.data_driven.discipline import (  # noqa:F811
        ScalableDiscipline,
    )

    return ScalableDiscipline(name, data, sizes, **parameters)


def create_surrogate(
    surrogate,  # type: Union[str,MLRegressionAlgo]
    data=None,  # type: Optional[Dataset]
    transformer=MLRegressionAlgo.DEFAULT_TRANSFORMER,  # type: Optional[TransformerType]
    disc_name=None,  # type: Optional[str]
    default_inputs=None,  # type: Optional[Dict[str,ndarray]]
    input_names=None,  # type: Optional[Iterable[str]]
    output_names=None,  # type: Optional[Iterable[str]]
    **parameters  # type: Any
):  # type: (...) -> SurrogateDiscipline
    """Create a surrogate discipline, either from a dataset or a regression model.

    Args:
        surrogate: Either the class name
            or the instance of the :class:`.MLRegressionAlgo`.
        data: The learning dataset to train the regression model.
            If None, the regression model is supposed to be trained.
        transformer: The strategies to transform the variables.
            The values are instances of :class:`.Transformer`
            while the keys are the names of
            either the variables
            or the groups of variables,
            e.g. "inputs" or "outputs" in the case of the regression algorithms.
            If a group is specified,
            the :class:`.Transformer` will be applied
            to all the variables of this group.
            If None, do not transform the variables.
            The :attr:`.MLRegressionAlgo.DEFAULT_TRANSFORMER` uses
            the :class:`.MinMaxScaler` strategy for both input and output variables.
        disc_name: The name to be given to the surrogate discipline.
            If None, concatenate :attr:`.ABBR` and ``data.name``.
        default_inputs: The default values of the inputs.
            If None, use the center of the learning input space.
        input_names: The names of the input variables.
            If None, consider all input variables mentioned in the learning dataset.
        output_names: The names of the output variables.
            If None, consider all input variables mentioned in the learning dataset.
        **parameters: The parameters of the machine learning algorithm.

    See also
    --------
    get_available_surrogates
    get_surrogate_options_schema
    """
    from gemseo.core.surrogate_disc import SurrogateDiscipline  # noqa:F811

    return SurrogateDiscipline(
        surrogate,
        data,
        transformer,
        disc_name,
        default_inputs,
        input_names,
        output_names,
        **parameters
    )


def create_mda(
    mda_name,  # type: str
    disciplines,  # type: Sequence[MDODiscipline]
    **options  # type: Any
):  # type: (...) -> MDA
    """Create an multidisciplinary analysis (MDA).

    Args:
        mda_name: The name of the MDA.
        disciplines: The disciplines.
        **options: The options of the MDA.

    Returns:
        The MDA.

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_mda
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2'])
    >>> mda = create_mda('MDAGaussSeidel', disciplines)
    >>> mda.execute()
    {'x_local': array([0.+0.j]),
     'x_shared': array([1.+0.j, 0.+0.j]),
     'y_0': array([0.79999995+0.j]),
     'y_1': array([1.79999995+0.j])}

    See also
    --------
    create_mda
    get_available_mdas
    get_mda_options_schema
    """
    from gemseo.mda.mda_factory import MDAFactory

    factory = MDAFactory()
    return factory.create(mda_name=mda_name, disciplines=disciplines, **options)


def execute_post(
    to_post_proc,  # type:Union[Scenario,OptimizationProblem,str,Path]
    post_name,  # type: str
    **options  # type: Any
):  # type: (...) -> Dict[str,Figure]
    """Post-process a result.

    Args:
        to_post_proc: The result to be post-processed,
            either a DOE scenario,
            a MDO scenario,
            an optimization problem
            or a path to an HDF file containing a saved optimization problem.
        post_name: The name of the post-processing.
        **options: The post-processing options.

    Returns:
        The figures, to be customized if not closed.

    Examples
    --------
    >>> from gemseo.api import create_discipline, create_scenario, execute_post
    >>> from gemseo.problems.sellar.sellar_design_space import SellarDesignSpace
    >>> disciplines = create_discipline(['Sellar1', 'Sellar2', 'SellarSystem'])
    >>> design_space = SellarDesignSpace()
    >>> scenario = create_scenario(disciplines, 'MDF', 'obj', design_space,
    >>>                            'SellarMDFScenario', 'MDO')
    >>> scenario.execute({'algo': 'NLOPT_SLSQP', 'max_iter': 100})
    >>> execute_post(scenario, 'OptHistoryView', show=False, save=True)

    See also
    --------
    get_available_post_processings
    get_post_processing_options_schema
    """
    from gemseo.algos.opt_problem import OptimizationProblem
    from gemseo.post.post_factory import PostFactory

    if hasattr(to_post_proc, "is_scenario") and to_post_proc.is_scenario():
        opt_problem = to_post_proc.formulation.opt_problem
    elif isinstance(to_post_proc, OptimizationProblem):
        opt_problem = to_post_proc
    elif isinstance(to_post_proc, string_types):
        opt_problem = OptimizationProblem.import_hdf(to_post_proc)
    else:
        raise TypeError("Cannot post process type: {}".format(type(to_post_proc)))
    return PostFactory().execute(opt_problem, post_name, **options)


def execute_algo(
    opt_problem,  # type: OptimizationProblem
    algo_name,  # type: str
    algo_type="opt",  # type: str
    **options  # type: Any
):  # type: (...) -> OptimizationResult
    """Solve an optimization problem.

    Args:
        opt_problem: The optimization problem to be solved.
        algo_name: The name of the algorithm to be used to solve optimization problem.
        algo_type: The type of algorithm,
            either "opt" for optimization
            or "doe" for design of experiments.
        **options: The options of the algorithm.

    Examples
    --------
    >>> from gemseo.api import execute_algo
    >>> from gemseo.problems.analytical.rosenbrock import Rosenbrock
    >>> opt_problem = Rosenbrock()
    >>> opt_result = execute_algo(opt_problem, 'SLSQP')
    >>> opt_result
    Optimization result:
    |_ Design variables: [0.99999787 0.99999581]
    |_ Objective function: 5.054173713127532e-12
    |_ Feasible solution: True

    See also
    --------
    get_available_opt_algorithms
    get_available_doe_algorithms
    get_algorithm_options_schema
    """
    if algo_type == "opt":
        from gemseo.algos.opt.opt_factory import OptimizersFactory

        factory = OptimizersFactory()

    elif algo_type == "doe":
        from gemseo.algos.doe.doe_factory import DOEFactory

        factory = DOEFactory()
    else:
        raise ValueError(
            "Unknown algo type: {}, please use 'doe' or 'opt' !".format(algo_type)
        )

    return factory.execute(opt_problem, algo_name, **options)


def monitor_scenario(
    scenario,  # type: Scenario
    observer,
):  # type: (...) -> None
    """Add an observer to a scenario.

    The observer must have an :meth:`update` method
    that handles the execution status change of an atom.
    `update(atom)` is called everytime an atom execution changes.

    Args:
        scenario: The scenario to monitor.
        observer: The observer that handles an update of status.

    See also
    --------
    create_scenario
    get_available_scenario_types
    get_scenario_options_schema
    get_scenario_inputs_schema
    get_scenario_differentiation_modes
    """
    from gemseo.core.monitoring import Monitoring

    # Monitoring object is a singleton
    monitor = Monitoring(scenario)
    monitor.add_observer(observer)


def print_configuration():  # type: (...) -> None
    """Print the current configuration.

    The log message contains the successfully loaded modules
    and failed imports with the reason.

    Examples
    --------
    >>> from gemseo.api import print_configuration
    >>> print_configuration()
    """
    from gemseo.algos.doe.doe_factory import DOEFactory
    from gemseo.algos.opt.opt_factory import OptimizersFactory
    from gemseo.formulations.formulations_factory import MDOFormulationsFactory
    from gemseo.mda.mda_factory import MDAFactory
    from gemseo.mlearning.regression.factory import RegressionModelFactory
    from gemseo.post.post_factory import PostFactory
    from gemseo.problems.disciplines_factory import DisciplinesFactory

    for factory in (
        DisciplinesFactory,
        OptimizersFactory,
        DOEFactory,
        RegressionModelFactory,
        MDOFormulationsFactory,
        MDAFactory,
        PostFactory,
    ):
        factory_repr = repr(factory().factory)
        LOGGER.info("%s", factory_repr)
        print(factory_repr)  # noqa: T001


def read_design_space(
    file_path,  # type: Union[str,Path]
    header=None,  # type: Optional[str]
):  # type: (...) -> DesignSpace
    """Read a design space from a file.

    Args:
        file_path: The path to the text file;
            it shall contain comma-separated values
            with a row for each variable
            and at least the bounds of the variable.
        header: The names of the fields saved in the file.
            If None, read them in the file.

    Returns:
        The design space.

    Examples
    --------
    >>> from gemseo.api import (create_design_space, export_design_space,
    >>>     read_design_space)
    >>> source_design_space = create_design_space()
    >>> source_design_space.add_variable('x', l_b=-1, value=0., u_b=1.)
    >>> export_design_space(source_design_space, 'file.txt')
    >>> read_design_space = read_design_space('file.txt')
    >>> print(read_design_space)
    Design Space:
    +------+-------------+-------+-------------+-------+
    | name | lower_bound | value | upper_bound | type  |
    +------+-------------+-------+-------------+-------+
    | x    |      -1     |   0   |      1      | float |
    +------+-------------+-------+-------------+-------+

    See also
    --------
    export_design_space
    create_design_space
    """
    from gemseo.algos.design_space import DesignSpace

    return DesignSpace.read_from_txt(file_path, header)


def export_design_space(
    design_space,  # type: DesignSpace
    output_file,  # type: Union[str,Path],
    export_hdf=False,  # type: bool
    fields=None,  # type: Optional[Sequence[str]]
    header_char="",  # type: str
    **table_options  # type: Any
):  # type: (...) -> None
    """Save a design space to a text or HDF file.

    Args:
        design_space: The design space to be saved.
        output_file: The path to the file.
        export_hdf: If True, save to an HDF file.
            Otherwise, save to a text file.
        fields: The fields to be exported.
            If None, export all fields.
        header_char: The header character.
        **table_options: The names and values of additional attributes
            for the :class:`.PrettyTable` view
            generated by :meth:`get_pretty_table`.

    Examples
    --------
    >>> from gemseo.api import create_design_space, export_design_space
    >>> design_space = create_design_space()
    >>> design_space.add_variable('x', l_b=-1, u_b=1, value=0.)
    >>> export_design_space(design_space, 'file.txt')

    See also
    --------
    read_design_space
    create_design_space
    """
    if export_hdf:
        design_space.export_hdf(output_file)
    else:
        design_space.export_to_txt(output_file, fields, header_char, **table_options)


def create_design_space():  # type: (...) -> DesignSpace
    """Create an empty design space.

    Returns:
        An empty design space.

    Examples
    --------
    >>> from gemseo.api import create_design_space
    >>> design_space = create_design_space()
    >>> design_space.add_variable('x', l_b=-1, u_b=1, value=0.)
    >>> print(design_space)
    Design Space:
    +------+-------------+-------+-------------+-------+
    | name | lower_bound | value | upper_bound | type  |
    +------+-------------+-------+-------------+-------+
    | x    |      -1     |   0   |      1      | float |
    +------+-------------+-------+-------------+-------+

    See also
    --------
    read_design_space
    export_design_space
    create_design_space
    gemseo.algos.design_space.DesignSpace
    """
    from gemseo.algos.design_space import DesignSpace

    return DesignSpace()


def create_parameter_space():  # type: (...) -> ParameterSpace
    """Create an empty parameter space.

    Returns:
        An empty parameter space.
    """
    from gemseo.algos.parameter_space import ParameterSpace  # noqa: F811

    return ParameterSpace()


def get_available_caches():  # type: (...) -> List[str]
    """Return the names of the available caches.

    Returns:
        The names of the available caches.

    Examples
    --------
    >>> from gemseo.api import get_available_caches
    >>> get_available_caches()
    ['AbstractFullCache', 'HDF5Cache', 'MemoryFullCache', 'SimpleCache']

    See also
    --------
    get_available_caches
    gemseo.core.discipline.MDODiscipline.set_cache_policy
    """
    from gemseo.caches.cache_factory import CacheFactory

    return CacheFactory().caches


def create_cache(
    cache_type,  # type: str
    name=None,  # type: Optional[str]
    **options  # type: Any
):  # type: (...) -> AbstractCache
    """Return a cache.

    Args:
        cache_type: The type of the cache.
        name: The name to be given to the cache.
            If None, use `cache_type`.
        **options: The options of the cache.

    Returns:
        The cache.

    Examples
    --------
    >>> from gemseo.api import create_cache
    >>> cache = create_cache('MemoryFullCache')
    >>> print(cache)
    +--------------------------------+
    |        MemoryFullCache         |
    +--------------+-----------------+
    |   Property   |      Value      |
    +--------------+-----------------+
    |     Type     | MemoryFullCache |
    |  Tolerance   |       0.0       |
    | Input names  |       None      |
    | Output names |       None      |
    |    Length    |        0        |
    +--------------+-----------------+

    See also
    --------
    get_available_caches
    gemseo.core.discipline.MDODiscipline.set_cache_policy
    """
    from gemseo.caches.cache_factory import CacheFactory

    return CacheFactory().create(cache_type, name=name, **options)


def create_dataset(
    name,  # type: str
    data,  # type: Union[ndarray,str,Path]
    variables=None,  # type: Optional[List[str]]
    sizes=None,  # type: Optional[Dict[str,int]]
    groups=None,  # type: Optional[Dict[str,str]]
    by_group=True,  # type: bool
    delimiter=",",  # type: str
    header=True,  # type: bool
    default_name=None,  # type: Optional[str]
):  # type: (...) -> Dataset
    """Create a dataset from a NumPy array or a data file.

    Args:
        name: The name to be given to the dataset.
        data: The data to be stored in the dataset,
            either a NumPy array or a file path.
        variables: The names of the variables.
            If None and `header` is True,
            read the names from the first line of the file.
            If None and `header` is False,
            use default names
            based on the patterns the :attr:`.Dataset.DEFAULT_NAMES`
            associated with the different groups.
        sizes: The sizes of the variables.
            If None,
            assume that all the variables have a size equal to 1.
        groups: The groups of the variables.
            If None,
            use :attr:`.Dataset.DEFAULT_GROUP` for all the variables.
        by_group: If True, store the data by group.
            Otherwise, store them by variables.
        delimiter: The field delimiter.
        header: If True and `data` is a string,
            read the variables names on the first line of the file.
        default_name: The name of the variable to be used as a pattern
            when variables is None.
            If None,
            use the :attr:`.Dataset.DEFAULT_NAMES` for this group if it exists.
            Otherwise, use the group name.

    Returns:
        The dataset generated from the NumPy array or data file.

    See also
    --------
    load_dataset
    """
    from gemseo.core.dataset import Dataset

    dataset = Dataset(name, by_group)
    if isinstance(data, ndarray):
        dataset.set_from_array(data, variables, sizes, groups, default_name)
    else:
        dataset.set_from_file(data, variables, sizes, groups, delimiter, header)
    return dataset


def load_dataset(
    dataset,  # type: str
    **options  # type: Any
):  # type: (...) -> Dataset
    """Instantiate a dataset.

    Typically, benchmark datasets can be found in :mod:`gemseo.problems.dataset`.

    Args:
        dataset: The name of the dataset (its class name).

    Returns:
        The dataset.

    See also
    --------
    create_dataset
    """
    from gemseo.problems.dataset.factory import DatasetFactory

    return DatasetFactory().create(dataset, **options)


def compute_doe(
    variables_space,  # type: DesignSpace
    algo_name,  # type: str
    size=None,  # type: Optional[int]
    unit_sampling=False,  # type: bool
    **options  # type: DOELibraryOptionType
):  # type: (...) -> ndarray
    """Compute a design of experiments (DOE) in a variables space.

    Args:
        variables_space: The variables space to be sampled.
        size: The size of the DOE.
            If ``None``, the size is deduced from the ``options``.
        algo_name: The DOE algorithm.
        unit_sampling: Whether to sample in the unit hypercube.
        **options: The options of the DOE algorithm.

    Returns:
          The design of experiments
          whose rows are the samples and columns the variables.

    Examples
    --------
    >>> from gemseo.api import compute_doe, create_design_space
    >>> variables_space = create_design_space()
    >>> variables_space.add_variable("x", 2, l_b=-1.0, u_b=1.0)
    >>> doe = compute_doe(variables_space, algo_name="lhs", size=5)

    See also
    --------
    get_available_doe_algorithms
    get_algorithm_options_schema
    execute_algo
    """
    library = DOEFactory().create(algo_name)
    return library.compute_doe(
        variables_space, size=size, unit_sampling=unit_sampling, **options
    )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Here is the baseclass to measure the quality of machine learning algorithms.

The concept of clustering quality measure is implemented with the
:class:`.MLClusteringMeasure` class and proposes different evaluation methods.
"""
from __future__ import division, unicode_literals

from copy import deepcopy
from typing import Dict, NoReturn, Optional, Sequence, Union

from numpy import arange
from numpy import delete as npdelete
from numpy import ndarray, unique
from numpy.random import choice

from gemseo.core.dataset import Dataset
from gemseo.mlearning.cluster.cluster import (
    MLClusteringAlgo,
    MLPredictiveClusteringAlgo,
)
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure


class MLClusteringMeasure(MLQualityMeasure):
    """An abstract clustering measure for clustering algorithms."""

    def __init__(
        self,
        algo,  # type: MLClusteringAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for clustering.
        """
        super(MLClusteringMeasure, self).__init__(algo)

    def evaluate_learn(
        self,
        samples=None,  # type: Optional[Sequence[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if not self.algo.is_trained:
            self.algo.learn(samples)
        data = self._get_data()[samples]
        labels = self.algo.labels
        measure = self._compute_measure(data, labels, multioutput)
        return measure

    def _compute_measure(
        self,
        data,  # type: ndarray
        labels,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        """Compute the quality measure.

        Args:
            data: The reference data.
            labels: The predicted labels.
            multioutput: Whether to return the quality measure
                for each output component; if not, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def _get_data(self):  # type: (...) -> Dict[str,ndarray]
        """Get data.

        Returns:
            The learning data indexed by the names of the variables.
        """
        names = self.algo.var_names
        data = self.algo.learning_set.get_data_by_names(names, False)
        return data


class MLPredictiveClusteringMeasure(MLClusteringMeasure):
    """An abstract clustering measure for predictive clustering algorithms."""

    def __init__(
        self,
        algo,  # type: MLPredictiveClusteringAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for predictive clustering.
        """
        super(MLPredictiveClusteringMeasure, self).__init__(algo)

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[Sequence[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if not self.algo.is_trained:
            self.algo.learn(samples)
        names = self.algo.var_names
        data = test_data.get_data_by_names(names, False)
        predictions = self.algo.predict(data)
        measure = self._compute_measure(data, predictions, multioutput)
        return measure

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[Sequence[int]]
        multioutput=True,  # type: bool
        randomize=False,  # type:bool
    ):  # type: (...) -> Union[float,ndarray]
        folds, samples = self._compute_folds(samples, n_folds, randomize)

        data = self._get_data()

        algo = deepcopy(self.algo)

        qualities = []
        for n_fold in range(n_folds):
            test_indices = folds[n_fold]
            train_indices = npdelete(samples, test_indices)
            algo.learn(samples=train_indices)
            test_data = data[test_indices]
            predictions = algo.predict(test_data)
            quality = self._compute_measure(test_data, predictions, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[Sequence[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        n_samples = samples.size
        indices = arange(n_samples)

        data = self._get_data()

        algo = deepcopy(self.algo)

        qualities = []
        for _ in range(n_replicates):
            train_indices = unique(choice(n_samples, n_samples))
            test_indices = npdelete(indices, train_indices)
            algo.learn(samples=[samples[index] for index in train_indices])
            test_data = data[[samples[index] for index in test_indices]]
            predictions = algo.predict(test_data)

            quality = self._compute_measure(test_data, predictions, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def _compute_measure(
        self,
        data,  # type: ndarray
        labels,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Compute the quality measure.

        Args:
            data: The reference data.
            labels: The predicted labels.
            multioutput: Whether to return the quality measure
                for each output component. If not, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def _get_data(self):  # type: (...) -> Dict[str,ndarray]
        """Get data.

        Returns:
            The learning data indexed by the names of the variables.
        """
        names = self.algo.var_names
        data = self.algo.learning_set.get_data_by_names(names, False)
        return data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""A Grammar based on JSON schema."""

from __future__ import division, unicode_literals

import json
import logging
from numbers import Number
from typing import (
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Union,
)

from numpy import generic, ndarray, zeros

from gemseo.core.grammars.abstract_grammar import AbstractGrammar
from gemseo.core.grammars.errors import InvalidDataException
from gemseo.core.grammars.json_schema import MutableMappingSchemaBuilder
from gemseo.core.grammars.simple_grammar import SimpleGrammar
from gemseo.utils.py23_compat import PY2, JsonSchemaException, Path, compile_schema
from gemseo.utils.string_tools import MultiLineString

if PY2:
    import jsonschema
    from jsonschema import ValidationError
else:
    # TODO: remove when py27 is gone
    class ValidationError(BaseException):
        pass


LOGGER = logging.getLogger(__name__)

ElementType = Union[str, float, bool, Sequence[Union[str, float, bool]]]
NumPyNestedMappingType = Mapping[
    str, Union[ElementType, ndarray, generic, "NumPyNestedMappingType"]
]
MappingSchemaType = Dict[
    str, Union[ElementType, List[ElementType], "MappingSchemaType"]
]
DictSchemaType = Mapping[str, Union[ElementType, List[ElementType], "DictSchemaType"]]
SerializedGrammarType = Dict[
    str, Union[ElementType, List[ElementType], "SerializedGrammarType"]
]


class JSONGrammar(AbstractGrammar):
    """A grammar based on a JSON schema.

    Attributes:
        schema (MutableMappingSchemaBuilder): The JSON schema.
    """

    PROPERTIES_FIELD = "properties"
    REQUIRED_FIELD = "required"
    TYPE_FIELD = "type"
    OBJECT_FIELD = "object"
    TYPES_MAP = {
        "array": ndarray,
        "float": float,
        "string": str,
        "integer": int,
        "boolean": bool,
        "number": Number,
    }

    def __init__(
        self,
        name,  # type: str
        schema_file=None,  # type: Optional[Union[str,Path]]
        schema=None,  # type: Optional[MappingSchemaType]
        descriptions=None,  # type: Optional[Mapping[str,str]]
    ):
        """
        Args:
            schema_file: The JSON schema file.
                If None, do not initialize the grammar from a JSON schema file.
            schema: A genson schema to initialize the grammar.
                If None,  do not initialize the grammar from a JSON schema.
            descriptions: The descriptions of the elements,
                in the form: ``{element_name: element_meaning}``.
                If None, use the descriptions available in the JSON schema if any.
        """
        super(JSONGrammar, self).__init__(name)
        self._validator = None
        self.schema = None
        self._schema_dict = None
        self._properties_dict = None
        self._init_schema()

        if schema is not None:
            self.schema.add_schema(schema)
        elif schema_file is not None:
            self.init_from_schema_file(schema_file, descriptions=descriptions)
        else:
            self.initialize_from_base_dict({}, description_dict=descriptions)

    def __repr__(self):  # type: (...) -> str
        return "{}, schema: {}".format(self, self.schema.to_json())

    def _init_schema(self):  # type: (...) -> None
        """Initialize the schema."""
        self.schema = MutableMappingSchemaBuilder()
        self._schema_dict = None
        self._properties_dict = None

    @property
    def schema_dict(self):  # type: (...) -> Dict[str,DictSchemaType]
        """The dictionary representation of the schema."""
        if self._schema_dict is None:
            self._schema_dict = self.schema.to_schema()
        return self._schema_dict

    @property
    def properties_dict(self):  # type: (...) -> Dict[str,DictSchemaType]
        """The dictionnary representation of the properties of the schema.

        Raises:
            ValueError: When the schema has no properties.
        """
        if self._properties_dict is None:
            self._properties_dict = self.schema_dict.get("properties")
            if self._properties_dict is None:
                raise ValueError(
                    "Schema has no properties: {}.".format(self.schema_dict)
                )
        return self._properties_dict

    def clear(self):  # type: (...) -> None
        self.__set_grammar_from_dict({})

    def _init_validator(self):  # type: (...) -> None
        """Initialize the validator."""
        self.schema_dict.pop("id", None)

        if PY2:
            # Use jsonschema instead of fastjsonschema when a property has anyOf.
            for value in self.schema_dict.get("properties", {}).values():
                if "anyOf" in value:
                    self._validator = jsonschema.validators.validator_for(
                        self.schema_dict
                    )(self.schema_dict).validate
                    return

        self._validator = compile_schema(self.schema_dict)

    @classmethod
    def cast_array_to_list(
        cls,
        data_dict,  # type: NumPyNestedMappingType
    ):  # type: (...) -> DictSchemaType
        """Cast the NumPy arrays to lists for dictionary values.

        Args:
            data_dict: The data mapping.

        Returns:
            The original mapping casted to a dictionary
            where NumPy arrays have been replaced with lists.
        """
        dict_of_list = dict(data_dict)
        for key, value in data_dict.items():
            if isinstance(value, (ndarray, generic)):
                dict_of_list[key] = value.real.tolist()
            elif isinstance(value, Mapping):
                dict_of_list[key] = cls.cast_array_to_list(value)

        return dict_of_list

    def is_required(
        self, element_name  # type: str
    ):  # type: (...) -> bool
        required_element_names = self.schema_dict.get("required", [element_name])
        return element_name in required_element_names

    def load_data(
        self,
        data,  # type: MutableMapping[str,ElementType]
        raise_exception=True,  # type: bool
    ):  # type: (...) -> MutableMapping[str,ElementType]
        """
        Raises:
            InvalidDataException:
                * If the passed data is not a dictionary.
                * If the data is not consistent with the grammar.
        """
        if not isinstance(data, MutableMapping):
            raise InvalidDataException(
                "Data must be a mutable mapping; "
                "got a {} instead.".format(type(data))
            )

        if self._validator is None:
            self._init_validator()

        data_to_check = self.cast_array_to_list(data)

        try:
            self._validator(data_to_check)
        except (JsonSchemaException, ValidationError) as error:
            log_message = MultiLineString()
            log_message.add("Invalid data in: {}".format(self.name))

            error_message = error.args[0]
            if error_message.startswith("data must contain"):
                # Error messages are not clear enough when missing elements
                # All keys are put in the message
                missing_elements = set(self.get_data_names()) - set(data.keys())

                if missing_elements:
                    log_message.add(
                        "Missing mandatory elements: {}".format(
                            ",".join(sorted(missing_elements))
                        )
                    )
                else:
                    log_message.add(", error: {}".format(error_message))
            else:
                log_message.add(", error: {}".format(error_message))

            LOGGER.error(log_message)

            if raise_exception:
                raise InvalidDataException(str(log_message))

            # Check a copy to keep types and arrays but store initial dict for complex
            # Add defaults
        for key, value in data_to_check.items():
            data.setdefault(key, value)

        return data

    def init_from_schema_file(
        self,
        schema_path,  # type: Union[str,Path]
        descriptions=None,  # type: Optional[Mapping[str,str]]
    ):  # type: (...) -> None
        """Set the grammar from a file.

        Args:
            schema_path: The path to the schema file.
            descriptions: The descriptions for the elements of the grammar,
                in the form: ``{element_name: element_meaning}``.
                If None, use the descriptions from the schema file.

        Raises:
            FileNotFoundError: If the schema file does not exist.
        """
        schema_path = Path(schema_path)

        if not schema_path.exists():
            raise FileNotFoundError(
                "Try to initialize grammar "
                "with not existing file: {}.".format(schema_path)
            )

        schema = json.loads(schema_path.read_text())
        self.__set_grammar_from_dict(schema, descriptions)

    def __set_grammar_from_dict(
        self,
        schema,  # type: Union[MappingSchemaType,MutableMappingSchemaBuilder]
        descriptions=None,  # type: Optional[Mapping[str,str]]
    ):  # type: (...) -> None
        """Set the grammar from a dictionary.

        Args:
            schema: The schema to set the grammar with.
            descriptions: The descriptions for the elements of the grammar,
                in the form: ``{element_name: element_meaning}``.
                If None, use the ``schema`` ones.
        """
        self._init_schema()
        self.__update_grammar_from_dict(schema, descriptions)

    def __update_grammar_from_dict(
        self,
        schema,  # type: Union[MappingSchemaType,MutableMappingSchemaBuilder]
        descriptions=None,  # type: Optional[Mapping[str,str]]
    ):  # type: (...) -> None
        """Update the grammar from a dictionary.

        Args:
            schema: The schema to update the grammar with.
            descriptions: The descriptions for the elements of the grammar,
                in the form: ``{element_name: element_meaning}``.
                If None, use the ``schema`` ones.
        """
        if descriptions is not None:
            if not isinstance(schema, dict):
                schema = schema.to_schema()

            for property_name, property_schema in schema["properties"].items():
                descr = descriptions.get(property_name)
                if descr is not None:
                    self.__add_description_to_types(descr, property_schema)

        self.__merge_schema(schema)

    def __add_description_to_types(
        self,
        description,  # type: str
        property_schema,  # type: Mapping[str, str]
    ):  # type: (...) -> None
        """Add the description for all the types found in the schema of a parameter.

        Args:
            description: The description of the parameter.
            property_schema: The schema of the parameter.
        """

        if "anyOf" in property_schema:
            for each_type in property_schema["anyOf"]:
                each_type["description"] = description
        else:
            property_schema["description"] = description

    def __merge_schema(
        self,
        schema,  # type: MappingSchemaType
    ):  # type: (...) -> None
        """Merge a schema in the current one.

        Args:
            schema: The schema to be merge, could be a schema object or a dictionary.
        """
        self.schema.add_schema(schema)
        self.__reset_schema_attrs()

    def initialize_from_data_names(
        self,
        data_names,  # type: Iterable[str]
        descriptions=None,  # type: Optional[Mapping[str,str]]
    ):  # type: (...) -> None
        """Initialize the grammar from the names and descriptions of the elements.

        Use float type.

        Args:
            descriptions: The descriptions of the elements,
                in the form: ``{element_name: element_meaning}``.
                If None, do not initialize the elements with descriptions.
        """
        element_value = zeros(1)
        elements_values = {element_name: element_value for element_name in data_names}
        self.initialize_from_base_dict(elements_values, description_dict=descriptions)

    def initialize_from_base_dict(
        self,
        typical_data_dict,  # type: Mapping[str,ElementType]
        description_dict=None,  # type: Optional[Mapping[str,str]]
    ):  # type: (...) -> None
        """Initialize the grammar with types and names from a typical data entry.

        The keys of the ``typical_data_dict`` are the names of the elements.
        The types of the values of the ``typical_data_dict`` will be converted
        to JSON Schema types and define the elements of the JSON Schema.

        Args:
            description_dict: The descriptions of the data names,
                in the form: ``{element_name: element_meaning}``.
                If None, do not initialize the elements with descriptions.
        """
        # Convert arrays to list as for check
        list_data_dict = self.cast_array_to_list(typical_data_dict)
        self.schema.add_object(list_data_dict)
        self.__set_grammar_from_dict(self.schema, description_dict)

    def get_data_names(self):  # type: (...) -> List[str]
        return list(self.schema.keys())

    def is_data_name_existing(
        self,
        data_name,  # type: str
    ):  # type: (...) -> bool
        return data_name in self.schema._properties

    def is_type_array(
        self, data_name  # type: str
    ):  # type: (...) -> bool
        if not self.is_data_name_existing(data_name):
            raise ValueError("{} is not in the grammar.".format(data_name))
        prop = self.properties_dict.get(data_name)
        return "array" == prop.get("type")

    def is_all_data_names_existing(
        self,
        data_names,  # type: Iterable[str]
    ):  # type: (...) -> bool
        properties = self.schema._properties
        for data_name in data_names:
            if data_name not in properties:
                return False
        return True

    def update_from(
        self,
        input_grammar,  # type: JSONGrammar
    ):  # type: (...) -> None
        """
        Raises:
            TypeError: If the passed grammar is not a JSONGrammar.
        """
        if not isinstance(input_grammar, JSONGrammar):
            msg = (
                "A {} is expected as input, but an object of type {} "
                "has been provided.".format(self.__class__, type(input_grammar))
            )
            raise TypeError(msg)

        self.__merge_schema(input_grammar.schema)

    def to_simple_grammar(self):  # type: (...) -> SimpleGrammar
        """Convert to the base :class:`.SimpleGrammar` type.

        Ignore the features of JSONGrammar that are not supported by SimpleGrammar.

        Returns:
            A :class:`.SimpleGrammar` equivalent to the current grammar.
        """
        grammar = SimpleGrammar(self.name)
        schema_dict = self.schema_dict
        properties = schema_dict.get(self.PROPERTIES_FIELD, {})

        names_to_types = {}
        for property_name, property_description in properties.items():
            property_json_type = property_description.get("type")
            if property_json_type not in self.TYPES_MAP:
                property_type = None
            else:
                property_type = self.TYPES_MAP[property_description["type"]]

            names_to_types[property_name] = property_type

            if property_json_type == "array" and "items" in property_description:
                property_json_sub_type = property_description["items"].get("type")
                if property_json_sub_type not in ["number", "integer", None]:
                    message = (
                        "Unsupported type '{}' in JSONGrammar '{}' "
                        "for property '{}' in conversion to simple grammar."
                    ).format(property_json_sub_type, self.name, property_name)
                    LOGGER.warning(message)

            for feature in ["minItems", "maxItems", "additionalItems", "contains"]:
                if feature in property_description:
                    message = (
                        "Unsupported feature '{}' in JSONGrammar '{}' "
                        "for property '{}' in conversion to simple grammar."
                    ).format(feature, self.name, property_name)
                    LOGGER.warning(message)

        grammar.update_elements(**names_to_types)

        required_data_names = self.schema_dict.get("required", [])
        grammar.update_required_elements(**dict.fromkeys(required_data_names, True))

        optional_data_names = set(self.schema_dict.get("properties", [])) - set(
            required_data_names
        )
        grammar.update_required_elements(**dict.fromkeys(optional_data_names, False))

        return grammar

    def update_from_if_not_in(
        self,
        input_grammar,  # type: JSONGrammar
        exclude_grammar,  # type: JSONGrammar
    ):  # type: (...) -> None
        if not (
            isinstance(input_grammar, self.__class__)
            and isinstance(exclude_grammar, self.__class__)
        ):
            msg = self._get_update_error_msg(self, input_grammar, exclude_grammar)
            raise TypeError(msg)

        schema = MutableMappingSchemaBuilder()
        schema.add_schema(input_grammar.schema)

        for element_name in exclude_grammar.schema.keys():
            try:
                del schema[element_name]
            except KeyError:
                pass

        self.__merge_schema(schema)

    def restrict_to(
        self,
        data_names,  # type: Sequence[str]
    ):  # type: (...) -> None
        for element_name in list(self.schema.keys()):
            if element_name not in data_names:
                del self.schema[element_name]
        self.__reset_schema_attrs()

    def remove_item(
        self,
        item_name,  # type: str
    ):  # type: (...) -> None
        del self.schema[item_name]
        self.__reset_schema_attrs()

    def __reset_schema_attrs(self):  # type: (...) -> None
        """Resets the validator, properties dict and schema dict conversions."""
        self._validator = None
        self._properties_dict = None
        self._schema_dict = None

    def set_item_value(
        self,
        item_name,  # type: str
        item_value,  # type: Dict[str,str]
    ):  # type: (...) -> None
        """Set the value of an element.

        Args:
            item_name: The name of the element.
            item_value: The value of the element.

        Raises:
            ValueError: If the item is not in the grammar.
        """
        if not self.is_data_name_existing(item_name):
            raise ValueError("Item {} not in grammar {}.".format(item_name, self.name))
        schema = self.schema_dict
        schema[self.PROPERTIES_FIELD][item_name] = item_value

        self.__set_grammar_from_dict(schema)

    def write_schema(
        self,
        path=None,  # type: Optional[Path,str]
    ):  # type: (...) -> None
        """Write the schema to a file.

        Args:
            path: The file path.
                If None,
                then write to a file named after the grammar and with .json extension.
        """
        if path is None:
            path = Path(self.name).with_suffix(".json")
        else:
            path = Path(path)

        schema_json = self.schema.to_json()

        if PY2:
            # workaround, see https://stackoverflow.com/a/36003774
            x = json.dumps(
                schema_json,
                ensure_ascii=False,
            )
            if isinstance(x, str):
                x = unicode(x, "UTF-8")  # noqa: F821
            path.write_text(x)
        else:
            json.dump(
                schema_json,
                path.open("w", encoding="utf-8"),
            )

    def __getstate__(self):  # type: (...) -> SerializedGrammarType
        """Used by pickle to define what to serialize.

        Returns:
            The dict to serialize.
        """
        deserialized_grammar = dict(self.__dict__)
        deserialized_grammar.pop("_validator")
        # genson schema cannot be pickled: use its dictionary representation
        deserialized_grammar["schema"] = self.schema_dict
        return deserialized_grammar

    def __setstate__(
        self,
        serialized_grammar,  # type: SerializedGrammarType
    ):  # type: (...) -> None
        """Used by pickle to define what to deserialize.

        Args:
            data_dict: update self dict from data_dict to deserialize.
        """
        self.__dict__.update(serialized_grammar)
        # genson schema cannot be pickled: use its dictionary representation
        self.__set_grammar_from_dict(serialized_grammar.pop("schema"))
